import os
import math
from pyrateshield import Observable, Logger, GLOBAL_LOG_LEVEL 
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

from PyQt5.QtGui import QIcon
from matplotlib.backends.backend_qt5agg \
    import FigureCanvasQTAgg, NavigationToolbar2QT

from matplotlib.figure import Figure
from matplotlib.backend_bases import MouseButton


from matplotlib.backend_bases import  _Mode
import qtawesome
from pyrateshield import labels

LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10

class PointClicker(Observable):
    EVENT_POINT_CLICKED = 'event_point_clicked'

    def __init__(self, mpl_controller=None):
        super().__init__()
        self.mpl_controller = mpl_controller
        mouse = self.mpl_controller.mouse
        mouse.connect(self, mouse.LEFT_CLICK_EVENT, self.get_position)
        self.mpl_controller.pick = False

    def get_position(self, position):
        self.mpl_controller.pick = True
        self.mpl_controller.mouse.disconnect(self)
        self.emit(self.EVENT_POINT_CLICKED, position)
        


class MovePointByMouse(Observable):    
    #line_format = {'color': 'black', 'linewidth': 2, 'linestyle': '--'}
    EVENT_FINISHED = 'event_finished'
    
    def __init__(self, mouse=None, model_item=None, 
                 vertex_index=None, hold=True):
        
        Observable.__init__(self, log_level=LOG_LEVEL)
        self.logger.debug('Initializing mouse interaction')
        self.model_item = model_item
        self.mouse = mouse
        self.vertex_index = vertex_index
        self.hold = hold
        self.start_move()
        
    
    def start_move(self):
        self.logger.debug('Start movement by mouse')
        self.mouse.connect(self, self.mouse.MOVE_EVENT, self.move)
        if self.hold:
            self.mouse.connect(self, self.mouse.RELEASE_EVENT, 
                               self.finished_move)
        else:
            self.mouse.connect(self, self.mouse.LEFT_CLICK_EVENT, 
                               self.finished_move)
    
    def move(self, position):
        if self.vertex_index is not None:
            self.model_item.set_vertex(self.vertex_index, position)
        else:
            self.model_item.position = position
        self.mouse.draw()
    
    def finished_move(self, _=None):
        self.mouse.disconnect(self)
        self.logger.debug('Finished mouse movement')
        self.emit(self.EVENT_FINISHED)
        




class Mouse(Observable):
    LEFT_CLICK_EVENT = 'left_click_event'
    RIGHT_CLICK_EVENT = 'right_click_event'
    LEFT_DBLCLICK_EVENT = 'left_dblclick_event'
    KEY_PRESS_EVENT = 'key_press_event'
    
    
    MOVE_EVENT = 'move_event'
    RELEASE_EVENT = 'release_event'
    _last_x = None
    _last_y = None
    
    delta = 10
   
    def __init__(self, canvas):
        super().__init__(log_level=Logger.LEVEL_INFO)
        self.canvas = canvas
        
        self.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.canvas.setFocus()

        #self.toolbar = toolbar
        self.inaxes = False
        self.infigure = False
        self.canvas.mpl_connect('button_press_event', self.mouse_click)
        self.canvas.mpl_connect('button_release_event', self.mouse_release)
        self.canvas.mpl_connect('motion_notify_event', self.mouse_move)
        self.canvas.mpl_connect('axes_enter_event', self.enter_axes)
        self.canvas.mpl_connect('axes_leave_event', self.leave_axes)
        self.canvas.mpl_connect('figure_enter_event', self.enter_figure)
        self.canvas.mpl_connect('figure_leave_event', self.leave_figure)
        self.canvas.mpl_connect('key_press_event', self.keypress)
        self.canvas.mpl_connect('key_release_event', self.keyrelease)
        
        self.shift_pressed = False
    
    def draw(self):
        self.canvas.draw_idle()
        
    def keypress(self, event):
        if event.key in ('shift',):
            self.shift_pressed = True
        self.emit(self.KEY_PRESS_EVENT, event.key)
        
    def keyrelease(self, event):
        if event.key in ('shift',):
            self.shift_pressed = False
        
        
    def enter_figure(self, event_data=None):
        self.infigure = True
        
    def leave_figure(self, event_data=None):
        self.infigure = False
        self.inaxes = False
        
    def enter_axes(self, event_data=None):
        self.inaxes = True
        self.infigure = True
        
    def leave_axes(self, event_data=None):
        self.inaxes = False
    
    def mouse_click(self, event):
        # if self.toolbar.mode in ('pan/zoom', 'zoom rect'):
        #     return
        
        if not event.inaxes: return
        x = event.xdata
        y = event.ydata

        self.click_position = (x, y)
        
        if event.button is MouseButton.LEFT:
            if event.dblclick:
                self.emit(self.LEFT_DBLCLICK_EVENT, [x, y])
            else:
                self.emit(self.LEFT_CLICK_EVENT, [x, y])
        if event.button is MouseButton.RIGHT:
            self.emit(self.RIGHT_CLICK_EVENT, [x, y])
            
    def mouse_move(self, event):
        if not event.inaxes: return
        
        x = event.xdata
        y = event.ydata
        
        
        if self.shift_pressed:
            if abs(y-self.click_position[1]) > abs(x-self.click_position[0]):
                x = self.click_position[0]
            else:
                y = self.click_position[1]
        
        
        if self._last_x is not None and self._last_y is not None:
            dx = (x - self._last_x)
            dy = (y - self._last_y)
            
            if math.sqrt(dx**2 + dy**2) < self.delta:
                return
            
        
                
            
        self._last_x = x
        self._lasy_y = y
        
        self.emit(self.MOVE_EVENT, [x, y])

    def mouse_release(self, event):
        x = event.xdata
        y = event.ydata
        self.emit(self.MOVE_EVENT, [x, y])
        self.emit(self.RELEASE_EVENT)
        self._last_x = None
        self._last_y = None
        
class NavigationToolbar(NavigationToolbar2QT, Observable):
    
    # only display the buttons we need
    EVENT_TOOLBUTTON_CLICK = 'event_toolbutton_click'
    
    
    pyrateshield_dict = {'source_nm': labels.SOURCES_NM,
                         'source_ct': labels.SOURCES_CT,
                         'source_xray': labels.SOURCES_XRAY,
                         'critical_point': labels.CRITICAL_POINTS,
                         'wall': labels.WALLS,
                         'move': labels.FLOORPLAN,
                         'calculator': 'calculator',
                         'dosemap_style': 'dosemap_style',
                         'isotopes': labels.ISOTOPES}
    
    toolitems = (
        
        #('Back', 'Back to previous view', 'back', 'back'), 
        #('Forward', 'Forward to next view', 'forward', 'forward'), 
        #(None, None, None, None), 
        (labels.PYSHIELD, 'Pyshield dosemap', 'icon', 'pyshield'),
        (labels.RADTRACER, 'Radtracer dosemap', 'icon', 'radtracer'),
        ('New', 'Start a new empty project', 'mdi.new-box', 'new'),
        ('Image', 'Load floorplan image from disk', 'mdi.file-image', 'load_image'),
        ('Floorplan', 'Show floorplan whithout dosemap', 'mdi.floor-plan', 'refresh'),
        
        (None, None, None, None), 
        ('Reset', 'Reset zoom/pan', 'mdi.arrow-expand', 'home'), 
        ('Pan', 'Left button pans, Right button zooms\nx/y fixes axis, CTRL fixes aspect', 'mdi.pan', 'pan'), 
        ('Zoom', 'Zoom to rectangle\nx/y fixes axis, CTRL fixes aspect', 'mdi.magnify-scan', 'zoom'), 
        ('Move', 'Move all items on floorplan', 'mdi.arrow-all', 'move'),
        #(None, None, None, None), 
        #('Subplots', 'Configure subplots', 'subplots', 'configure_subplots'), 
        #(None, None, None, None), 
        
        
        (None, None, None, None), 
        ('Source NM', 'Add NM source by mouse click', 'mdi.radioactive', 'source_nm'),
        ('Source CT', 'Add CT source by mouse click', 'mdi.radiology-box-outline', 'source_ct'),
        ('Source Xray', 'Add Xray source by mouse click', 'mdi.radiology-box-outline', 'source_xray'),
        ('Wall', 'Draw new wall with mouse', 'mdi.wall', 'wall'),
        ('Critical Point', 'Add critical point by mouse click', 'mdi.map-marker-alert-outline', 'critical_point'),
        
        
        ('Delete', 'Delete selected item', 'mdi.trash-can-outline', 'delete'),
        (None, None, None, None), 
        
        #('Side Panel', 'Show/hide side panel', 'mdi.file-document-outline', 'side_panel'),
        
        

        
        ('Snapshot', 'Save current view as image', 'mdi.camera-outline', 'save_figure'),
        ('Load', 'Load project from disk', 'mdi.folder-open-outline', 'load_project'),
        ('Save', 'Save project', 'mdi.content-save-outline', 'save_project'),
        ('Save As', 'Save project as ...', 'mdi.content-save-edit-outline', 'save_project_as'),
        (None, None, None, None), 
        ('Calculator', 'Calculate Transmission for a shielding', 'mdi.calculator', 'calculator'),
        ('Dosemap style', 'Edit style setting for the dose map display', 'mdi.palette', 'dosemap_style'),
        ('Isotopes', 'View isotopes implemented by pyrateshield', 'mdi.atom', 'isotopes')
       
    )
    
    def __init__(self, *args, mpl_controller=None, **kwargs):
        NavigationToolbar2QT.__init__(self, *args, coordinates=False, **kwargs)
        Observable.__init__(self)
        self.setStyleSheet("border-bottom: 0px; border-top: 0px;")
        
        self.mpl_controller = mpl_controller
        
        
        for name, action in self._actions.items():
            widget = self.widgetForAction(action)
            
            if name in ('source_nm', 'source_ct', 'source_xray',
                        'critical_point', 'wall', 'zoom', 'pan', 'side_panel',
                        'move'):
           
                action.setCheckable(True)
                if name == 'side_panel':
                    action.setChecked(True)
                if name == 'save_project':
                    action.setEnabled(False)
                
                style = 'QToolButton:checked { background-color: darkgray }'
            else:
                style = 'QToolButton:pressed { background-color: darkgray }'
            widget.setStyleSheet(style)
            
            #if name in ('pyshield', 'radtracer', 'source_nm', 'source_ct', 'source_xray'):
            
            widget.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon);
            
        
    def _icon(self, name):
        if name == 'icon.png':
            folder = os.path.split(__file__)[0]
            icon = QIcon(os.path.join(folder, 'icon.png'))
        else:
            try:
                # .png is added remove last 4 characters
                icon = qtawesome.icon(name[:-4])
            except:
                icon = NavigationToolbar2QT._icon(self, name)
        return icon
    
    def button_enabled(self, label, enabled):
        if label in self.pyrateshield_dict.values():
            label = dict([(value, key)\
                          for key, value in self.pyrateshield_dict.items()])[label]
        
        self._actions[label].setEnabled(enabled)
    
    
    def button_checked(self, label, enabled):
        if label in self.pyrateshield_dict.values():
            label = dict([(value, key)\
                          for key, value in self.pyrateshield_dict.items()])[label]
        
        self._actions[label].setChecked(enabled)
    
        
    
    def delete(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('delete', False))
    
    def refresh(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('refresh', False))
        
        
    def load_project(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('load_project', False))
        
    def save_project_as(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('save_project_as', False))
        
    def new(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('new', False))
        
    def save_project(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('save_project', False))
        
    def load_image(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('load_image', False))
        
    def radtracer(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK, (labels.RADTRACER, False))
        
    def pyshield(self):
        self.emit(self.EVENT_TOOLBUTTON_CLICK,  (labels.PYSHIELD, False))
    
    @property
    def selected_tool(self):
        selected_item = None
        for name, action in self._actions.items():
            if name == 'side_panel':
                continue
            
            if action.isChecked():
                if selected_item is not None:
                    raise RuntimeError() # debug purpuses
                elif name in ('zoom', 'pan'):
                    return name
                else:
                    selected_item = self.pyrateshield_dict[name]
        return selected_item
    
    def tool_selected(self, toolname):
        checked = self._actions[toolname].isChecked()
        self.deselect_zoom_pan()
        if checked:
            self.select_checkable_tool(toolname)
        label = self.pyrateshield_dict[toolname]
        self.emit(self.EVENT_TOOLBUTTON_CLICK, (label, checked))
        
    def source_nm(self):
        self.tool_selected('source_nm')
        
    def source_ct(self):
        self.tool_selected('source_ct')
        
    def source_xray(self):
        self.tool_selected('source_xray')
        
    def wall(self):
        self.tool_selected('wall')
        
    def critical_point(self):
        self.tool_selected('critical_point')
        
    def calculator(self):
        self.tool_selected('calculator')
    
    def dosemap_style(self):
        self.tool_selected('dosemap_style')
    
    def isotopes(self):
        self.tool_selected('isotopes')
        
    def side_panel(self):
        checked = self._actions['side_panel'].isChecked()
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('side_panel', checked))
        
    def zoom(self):
        NavigationToolbar2QT.zoom(self)
        if self.mode == _Mode.ZOOM:
            self.deselect_model_items()
        
    def pan(self):
        NavigationToolbar2QT.pan(self)
        if self.mode == _Mode.PAN:
            self.deselect_model_items()
            
    def move(self):
        checked = self._actions['move'].isChecked()
        if checked:            
            QApplication.setOverrideCursor(Qt.SizeAllCursor)

        else:
            QApplication.restoreOverrideCursor()
        
        self.emit(self.EVENT_TOOLBUTTON_CLICK, ('move', checked))
        
    def select_checkable_tool(self, toolname=None):
        self.logger.debug('Select Tool')
        for name in ('source_nm', 'source_ct', 'source_xray',
                     'wall', 'critical_point', 'move'):
            if toolname != name:
                self._actions[name].setChecked(False)
                
       
    
    @property
    def zoom_pan_selected(self):
        return self.mode in (_Mode.PAN, _Mode.ZOOM)
    
    def deselect_zoom_pan(self):
        if self.mode == _Mode.PAN:
            self.pan()
        elif self.mode == _Mode.ZOOM:
            self.zoom()
    
    def deselect_model_items(self):
        for action in ('source_ct', 'source_xray', 'source_nm',
                       'critical_point', 'wall', 'move'):
            self._actions[action].setChecked(False)
            
            
    def deselect(self):
        self.deselect_zoom_pan()
        self.deselect_model_items()
        
    

class MplCanvas(FigureCanvasQTAgg):
    def __init__(self):
        self.fig = Figure()
        self.fig.set_facecolor('0.8') #Gray background around axes
        self.ax = self.fig.add_axes([0, 0, 1, 1])

        FigureCanvasQTAgg.__init__(self, self.fig)

        FigureCanvasQTAgg.setSizePolicy(self,
                                   QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Expanding)

        FigureCanvasQTAgg.updateGeometry(self)

    




import numpy as np
from pyrateshield.constants import CONSTANTS
from matplotlib.path import Path



def unit_vector(vector):
    return vector / np.linalg.norm(vector)


def get_polygon(point, line, extent):
    p1 = np.asarray(line[0])
    p2 = np.asarray(line[1])

    uv1 = unit_vector(p1-point)
    uv2 = unit_vector(p2-point)

    angle = np.arccos(np.clip(np.dot(uv1, uv2), -1.0, 1.0))

    min_dist = np.sqrt((extent[3] - extent[2]) ** 2\
                      + (extent[1] - extent[0]) **2)


    d = min_dist/np.cos(0.5 * angle)

    p3 = point + d*uv1
    p4 = point + d*uv2

    poly = np.array((p1, p3, p4, p2))
    return poly


def grid_points_in_polygon(grid, poly):

    points = np.asarray([grid.X.flatten(), grid.Y.flatten()]).T

    path = Path(poly)

    inside =  path.contains_points(points)

    mask = inside.reshape(grid.X.shape)

    return points[inside], mask


def get_intersection_map(points, point, vertices, mask):
    intersection_thickness = get_diagonal_intersection_factor(points.T, point, vertices)
    intersection_map = np.zeros(mask.shape)
    intersection_map[mask] = intersection_thickness
    return intersection_map
    

# line intersections
def ccw(A,B,C):
    # Is counter-clockwise?
    return (C[1]-A[1]) * (B[0]-A[0]) > (B[1]-A[1]) * (C[0]-A[0])

def intersect(A,B,C,D):
    # Return true if line segments AB and CD intersect
    return ccw(A,C,D) != ccw(B,C,D) and ccw(A,B,C) != ccw(A,B,D)

def get_diagonal_intersection_factor(poly, source_ji, vertices_ji):
    """ 
    Get diagonal distance factor through wall for every point within polygon (head-on=1, 45deg=sqrt(2))
    """
    r = (poly[0]-source_ji[0], poly[1]-source_ji[1]) # Vectors from source to points in polygon
    w = (vertices_ji[1][1]-vertices_ji[0][1], -vertices_ji[1][0]+vertices_ji[0][0]) # Vector orthogonal to wall (vec=v1-v0, w=(vec_y,-vec_x))
    r_in_w = r[0]*w[0] + r[1]*w[1] # Inner product for every vector in r with w
    mag_r_w = np.linalg.norm(r, axis=0) * (w[0]**2 + w[1]**2)**0.5 # Multiplied magnitude of vectors
    return np.abs(mag_r_w / r_in_w) # 1 / Cos(th)     ( cos(th) = u.v / (|u| |v|) )



def effective_decay_constant(physical_half_life=None, biological_half_life=None):
    
    effective_decay_constant = 0
    
    if physical_half_life is not None and physical_half_life > 0:
        effective_decay_constant += np.log(2) / physical_half_life
        
    if biological_half_life is not None and biological_half_life > 0:
        effective_decay_constant += np.log(2) / biological_half_life
        
    return effective_decay_constant


def remainder(clearance_model):
    # remaining fraction decays with physical_decay
    remainder = 1
    if clearance_model.apply_fraction1:
        remainder -= clearance_model.fraction1
    if clearance_model.apply_fraction2:
        remainder -= clearance_model.fraction2
    
    return remainder


def _tiac(activity, decay_constant, start=0, end=float('inf')):
    
    if decay_constant > 0:
        tiac = activity / decay_constant * (np.exp(-decay_constant * start)\
                                            - np.exp(-decay_constant * end))
    else:
        tiac = activity * (end - start)
        
        
    return tiac
            
def tiac(source, clearance_model=None):
    return time_integrated_activity_coefficient_mbqh(source, 
                                                     clearance_model=clearance_model)


def time_integrated_activity_coefficient_mbqh(source, clearance_model=None):
    isotope = CONSTANTS.get_isotope_by_name(source.isotope)
    
    if source.apply_decay_correction:
        physical_half_life = isotope.half_life
    else:
        physical_half_life = float('inf')

    if clearance_model is not None:
        # check if one of the fractions is checked
        apply_clearance_model = clearance_model.apply_fraction1\
            or clearance_model.apply_fraction2
    else:
        apply_clearance_model = False
    
    if not apply_clearance_model:
        # just simple physical decay
        end = source.duration
        tiac =  _tiac(source.activity, np.log(2) / physical_half_life, end=end)
        
    else:
        
        tiac = 0
        
        if clearance_model.apply_fraction1:
            # first fraction 1
            decay_constant1 = effective_decay_constant(physical_half_life,
                                                      clearance_model.half_life1)
            
            # check if fractions are split in time. If split determine end
            # time either by the split time or by the duration
            if clearance_model.apply_split_fractions:                
                end1 = min(clearance_model.split_time, source.duration)
            else:
                end1 = source.duration
            
            # fraction of activity to consider
            activity = source.activity * clearance_model.fraction1
            
            # calculate tiac for the considered fraction of activity
            tiac += _tiac(activity, decay_constant1, end=end1) 
            
        if clearance_model.apply_fraction2:
            if clearance_model.apply_split_fractions\
                and clearance_model.apply_fraction1:
            
                # If split in time determine remaining activity after decay
                # of fraction 1
                activity = source.activity * np.exp(-decay_constant1 * end1)
                
                start = clearance_model.split_time
                
            else:
                # if not split in time integrate from 0
                activity = source.activity
                start = 0 
            
            # fraction of activity to consider
            activity *= clearance_model.fraction2
            
            
            decay_constant2 = effective_decay_constant(physical_half_life,
                                                       clearance_model.half_life2)
            # always integrate to the source duration for fraction 2
            end = source.duration
            
            if start >= end:
                pass
                # if start >= end the start time is later than source duration
                # tiac = 0 for fraction 2
            else:                
                tiac += _tiac(activity, decay_constant2, start=start, end=end) 
        
        
        # determine remaining fraction
        remaining_fraction = 1
        
        if clearance_model.apply_fraction1:
            remaining_fraction -= clearance_model.fraction1

        if clearance_model.apply_fraction2:
            remaining_fraction -= clearance_model.fraction2
        
        # not that remaining fraction == -2 when fractions are split in time
        if remaining_fraction > 0:             
            # activity that is remaining
            activity = source.activity * remaining_fraction
            
            # integrate from 0 to source duration for remaining fraction
            tiac += _tiac(activity,
                          np.log(2) / physical_half_life,
                          end=source.duration)
        
            
    tiac *= source.number_of_exams
    
    return tiac
            
    
    

    


import inspect
from pyrateshield import labels, Observable, GLOBAL_LOG_LEVEL

LOG_LEVEL = GLOBAL_LOG_LEVEL

#LOG_LEVEL = 10

def equal_position(position1, position2):
    if  [position1, position2].count(None) == 1:
        return False
    elif [position1, position2].count(None) == 2:
        return True
    else:
        return position1[0] == position2[0]\
            and position1[1] == position2[1]
        
def equal_vertices(vertices1, vertices2):
    if  [vertices1, vertices2].count(None) == 1:
        return False
    elif [vertices1, vertices2].count(None) == 2:
        return True
    else:
        return equal_position(vertices1[0], vertices2[0])\
            and equal_position(vertices1[1], vertices2[1])

def equal_materials(materials1, materials2):
    if  [materials1, materials2].count(None) == 1:
        return False
    elif [materials1, materials2].count(None) == 2:
        return True
    elif len(materials1) != len(materials2):
            return False
    else:
        eq = True
        for material1, material2 in zip(materials1, materials2):
            eq = eq and material1[0] == material2[0] and material1[1] == material2[1]
        return eq
    

### Base class
class ModelItem(Observable):
    label = 'ModelItem'
    EVENT_UPDATE = 'event_update'
    EVENT_DELETE = 'event_delete'
    
    _attr_dct = {}
    _attr_defaults = {}   


    def __init__(self, **kwargs):        
        
        Observable.__init__(self, log_level=LOG_LEVEL)
        
        
        init_values = self._get_default_values()
        init_values.update(kwargs)
        self.logger.debug(f'Init with kwargs {str(init_values)}')
        for attr_name, value in init_values.items():
            setattr(self, attr_name, value)
            
    def delete(self):
        self.emit(self.EVENT_DELETE, self)
        self.disconnect()
        # will be picked up by garbage collector now
            
    def __copy__(self):
        return self.__class__.from_dict(self.to_dict())

    
    @classmethod
    def _get_default_value(cls, attr_name):
        if cls._attr_dct[attr_name] in cls._attr_defaults.keys():     
            # default value defined
            value = cls._attr_defaults[cls._attr_dct[attr_name]]

        elif attr_name in cls._attr_dct.keys(): 
            # No default value but is in the attribute dct
            value = None
        else:
            # Debug purpuses should not happen for a well defined ModelItem
            raise AttributeError(f'Unknown Attribute {attr_name}')
            
        if inspect.isclass(value) or inspect.isfunction(value):
            value = value() # call or create instance
        return value
    
    @classmethod
    def _get_default_values(self):
        values = {}
        for attr_name in self._attr_dct.keys():
            values[attr_name] = self._get_default_value(attr_name)
        return values
        
    def __setattr__(self, attr_name, value):
        
        if attr_name not in self._attr_dct.keys():
            return super().__setattr__(attr_name, value)
        
        
        if hasattr(self, attr_name):
            old_value = getattr(self, attr_name)
        else: 
            super().__setattr__(attr_name, None)
            old_value = None
        self.logger.debug(f'Setting Attribute {attr_name} from {str(old_value)} to {str(value)}')
        # Override None values to default values                
        if value is None:            
            value = self._get_default_value(attr_name)
            self.logger.debug(f'Overriding value with defualt {str(value)}')
        
        if attr_name == 'materials' and equal_materials(old_value, value):
                return # don't set and don't emit event
                
        elif attr_name == 'position':
            value = [float(value[0]), float(value[1])] # force list!
            if equal_position(old_value, value):
                return # don't set and don't emit event
                
        elif attr_name == 'vertices' :# force list!            
            value = [[float(value[0][0]), float(value[0][1])],
                     [float(value[1][0]), float(value[1][1])]]
                
            if equal_vertices(old_value, value):
                return # don't set and don't emit event
        elif attr_name == 'image': 
            if old_value is value:
                return
        else:
            if old_value == value:
                return  # don't set and don't emit event
    
        super().__setattr__(attr_name, value)
            
        if attr_name in self._attr_dct.keys():
            # gather event_data and emit event
            value = getattr(self, attr_name)
            label = self._attr_dct[attr_name]
            event_data = self, label, old_value, value
            self.emit(self.EVENT_UPDATE, event_data=event_data)
            
    def update_by_dict(self, dct):
        for key, value in dct.items():
            attr_name = self.attr_name_from_label(key)
            setattr(self, attr_name, value)
          
    @classmethod
    def from_dict(cls, dct):
        kwargs = {}
        for label, value in dct.items():
            if label in cls._attr_dct.values():
                kwargs[cls.attr_name_from_label(label)] = value
            elif label == labels.ENABLED or label == 'enable' or label=='show':
                # Tricky part here 
                # Some older psp / zip files contain model_items with the
                # enabled tag while it is not used. This would now error in 
                # code. Ignore the enabled flag below to ignore
                continue
            else:
                raise KeyError(f'{label} not a valid setting for {cls.__name__}')
                
            
        return cls(**kwargs)
    
    def enable(self, value):
        if labels.ENABLED in self._attr_dct.values():
            self.enabled = value
        else:
            raise AttributeError

    def to_dict(self):
        dct = {label: getattr(self, var)\
               for var, label in self._attr_dct.items()}
        return dct
    
    @classmethod
    def attr_name_from_label(cls, label):        
        index = list(cls._attr_dct.values()).index(label)
        return list(cls._attr_dct.keys())[index]
    
    
    def __str__(self):
        return "\n".join([f"{k}: {v}" for k, v in self.to_dict().items()])
    
    def __repr__(self):
        return self.__str__()
    
    





from pyrateshield import Logger, GLOBAL_LOG_LEVEL
LOG_LEVEL = GLOBAL_LOG_LEVEL

#LOG_LEVEL = 10

class Observable(Logger):
    
    """
    Superclass to generate events that can be observed externally.
    Used for a Model, View, Controller (MVC) GUI design
    """
    def __init__(self, log_level=None):
        if log_level is None:
            log_level = LOG_LEVEL
        super().__init__(log_level=log_level)
        
        self.callbacks = {}
        
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['callbacks']
        return state 

    def connect(self, listener, event_name, callback):
        """
        Add a listener for a specific event.

        Parameters
        ----------
        listener : Object
            Object that listens for event.
        event_name : string
            Name of event that is listent to.
        callback : function
            Function that is called when event is emitted/generated.
            A single positional argument is passed that contais specific data
            for the event.

        Returns
        -------
        None.

        """
        if self.log_level == Logger.LEVEL_DEBUG:
            if listener is not None:
                name_listener = listener.__class__.__name__
            else:
                name_listener = str(None)
            logstr = (f'Adding {name_listener} to event {event_name} of class '
                      f'{self.__class__.__name__} with callback '
                      f'{callback.__name__}')
            
            self.logger.debug(str(logstr))
            
        
        
        if listener not in self.callbacks.keys():
            self.callbacks[listener] = {}
        if event_name not in self.callbacks[listener].keys():
            self.callbacks[listener][event_name] = [True, callback]
        else:
            # connecting the same listener multiple times for the same callback
            # can lead to hard to debug errors
            # if LOG_LEVEL == Logger.LEVEL_DEBUG:
            #     raise KeyError(f'Listener already connected for {event_name}')
            # else:
            self.logger.debug(f'Listener already connected for {event_name}')

    def disconnect(self, listener=None, event_name=None):
        """
        Remove a specific listener/listener

        Parameters
        ----------
        listener : Object
            Object that listens for event.
        event_name : string, optional
            DIf specified only a specific event is removed for this listener.
            If None all events are unsubscribed. The default is None.

        Returns
        -------
        None.

        """
        # if len(self.callbacks) > 0:
        #     raise ValueError()
        if listener is None:
            for listener in self.callbacks.copy().keys():
                self.disconnect(listener=listener, event_name=event_name)
            return
        elif listener not in self.callbacks.keys():
            self.logger.debug(f"listener {listener.__class__.__name__} not connected")
            return
            
        elif event_name is None:
            for event_name in self.callbacks[listener].copy().keys():
                self.disconnect(listener=listener, event_name=event_name)

            return
        
        if listener not in self.callbacks.keys():
            return
        
        callback = self.callbacks[listener].pop(event_name)[1]
        
        if self.log_level == Logger.LEVEL_DEBUG:
            if listener is not None:
                name_listener = listener.__class__
            else:
                name_listener = str(None)
    
            
    
            logstr = f'Removing {name_listener} to event {event_name} of class {self.__class__.__name__} with callback {callback.__name__}'
            
            
            
            self.logger.debug(logstr)
        
        
            
                    

    def disable_connection(self, listener=None, event_name=None):
        """
        Disable a listener for events (temporarily).

        Parameters
        ----------
        listener : Object
            Object that listens for event.
        event_name : string, optional
            DIf specified only a specific event is disabled for this listener.
            If None all events are disabled. The default is None.

        Returns
        -------
        None.

        """
        if listener is None:
            for listener in self.callbacks.copy().keys():
                self.disable_connection(listener=listener, 
                                        event_name=event_name)

            return
            
        elif listener not in self.callbacks.keys():
            return
        
        if event_name is None:
            for event_name in self.callbacks[listener].copy().keys():
                self.disable_connection(listener=listener, 
                                        event_name=event_name)
 
            return

        if event_name in self.callbacks[listener].keys():
            self.callbacks[listener][event_name][0] = False
    
        if self.log_level == Logger.LEVEL_DEBUG:
        
            if listener is not None:
                name_listener = listener.__class__.__name__
            else:
                name_listener = str(None)
                    
            callback = self.callbacks[listener][event_name][1]
        
            logstr = f'Disabling {name_listener} to event {event_name} of class {self.__class__.__name__} with callback {callback.__name__}'        
            self.logger.debug(logstr)


    def enable_connection(self, listener=None, event_name=None):
        """
        Enable a listener for events (temporarily).

        Parameters
        ----------
        listener : Object
            Object that listens for event.
        event_name : string, optional
            DIf specified only a specific event is enabled for this listener.
            If None all events are enabled. The default is None.

        Returns
        -------
        None.

        """
        if listener is None:
            for listener in self.callbacks.copy().keys():
             
                self.enable_connection(listener=listener, 
                                        event_name=event_name)
            return
        
        elif listener not in self.callbacks.keys():
            return
            
        if event_name is None:
            for event_name in self.callbacks[listener].keys():
            
                self.enable_connection(listener=listener, 
                                        event_name=event_name)
            return
        
        if event_name in self.callbacks[listener].keys():
            self.callbacks[listener][event_name][0] = True
    
        if self.log_level == Logger.LEVEL_DEBUG:
        
            if listener is not None:
                name_listener = listener.__class__.__name__
            else:
                name_listener = str(None)
                    
            callback = self.callbacks[listener][event_name][1]
        
            logstr = f'Enabling {name_listener} to event {event_name} of class {self.__class__.__name__} with callback {callback.__name__}' 
            
            self.logger.debug(logstr)

    def emit(self, event_name, event_data=None):
        """
        Generate/Emit an event

        Parameters
        ----------
        event_name : string
            Name of the event that is generated/emitted.
        event_data : Object, optional
            Event data that is passed to listeners/listeners.
            The default is None.

        Returns
        -------
        None.

        """
        
        # mouse interactions add and remove callbacks quickly create copy
        # to prevent dictionary changes during iteration
        
        for listener, events in self.callbacks.copy().items():
            if event_name in events.keys():
                enabled, callback = self.callbacks[listener][event_name]
                

                if self.log_level == self.LEVEL_DEBUG:
                    if listener is not None:
                        name_listener = listener.__class__.__name__
                    else:
                        name_listener = str(None)
                

                    if enabled:
                        logstr = (f'Event {event_name} emitted, callback {callback.__name__} of {name_listener} is called') 
                    else:
                        logstr = (f'Event {event_name} emitted but disabled callback {callback.__name__} of {name_listener}') 
                        
                    self.logger.debug(logstr)
                if enabled:
                    callback(event_data)
                    
                    




import numpy as np

from pyrateshield import labels
from pyrateshield.modelitem import ModelItem

from pyrateshield.constants import CONSTANTS


def ZERO_VERTICES():
    return [[0, 0], [0, 0]]

def ZERO_POSITION():
   return [0, 0]      

def EMPTY_MATERIALS():
    # default for shielding
    return [[labels.EMPTY_MATERIAL, 0], 
            [labels.EMPTY_MATERIAL, 0]]

class Material(ModelItem):
    # buildup table and attenuation table are used by pyshield only
    default_name = 'Material'
    label = labels.MATERIALS
    _attr_dct = {
        "name": labels.NAME,
        "density": labels.DENSITY,
        "attenuation_table": labels.ATTENUATION_TABLE,
        "buildup_table": labels.BUILDUP_TABLE,
        "radtracer_material": labels.RADTRACER_MATERIAL
    }
    
    _attr_defaults = {labels.NAME: default_name,
                      labels.DENSITY: 1,
                      labels.ATTENUATION_TABLE: labels.EMPTY_TABLE,
                      labels.BUILDUP_TABLE: labels.EMPTY_TABLE,
                      labels.RADTRACER_MATERIAL: labels.EMPTY_MATERIAL}
                      
    
class Shielding(ModelItem):
    default_name = 'Shielding'
    label = labels.SHIELDINGS

    _attr_dct = {
        "name": labels.NAME,
        "color": labels.COLOR,
        "linewidth": labels.LINEWIDTH,
        "materials": labels.MATERIALS,
        
    }
    
    _attr_defaults = {labels.NAME: default_name,
                      labels.COLOR: 'black',
                      labels.LINEWIDTH: 2,
                      labels.MATERIALS: EMPTY_MATERIALS}
                      
   
    
    
class Clearance(ModelItem):
    default_name = 'Clearance Model'
    label = labels.CLEARANCE
    
    _attr_dct = {
        "name": labels.NAME,
        'apply_fraction1': labels.APPLY_FRACTION1,
        'apply_fraction2': labels.APPLY_FRACTION2,
        "fraction1": labels.DECAY_FRACTION1,
        "fraction2": labels.DECAY_FRACTION2,
        "half_life1": labels.HALFLIFE1,
        'half_life2': labels.HALFLIFE2,    
        'apply_split_fractions': labels.ENABLE_SPLIT_FRACTIONS,
        'split_time': labels.SPLIT_FRACTION_TIME}
       
    
    
    _attr_defaults = {labels.NAME: default_name,
                      labels.APPLY_FRACTION1: False,
                      labels.DECAY_FRACTION1: 0,
                      labels.HALFLIFE2: 24,
                      labels.APPLY_FRACTION2: False,
                      labels.DECAY_FRACTION2: 0,
                      labels.HALFLIFE1: 24,          
                      labels.ENABLE_SPLIT_FRACTIONS: False,
                      labels.SPLIT_FRACTION_TIME: 24}
    
            
    
    def is_equal(self, other):
        if isinstance(other, Clearance):
            eq = True
            for attr in self._attr_dct.keys():
                if attr == 'name': continue
                eq = eq and (getattr(self, attr) == getattr(other, attr))
            return eq
        else:
            return False
        
    
    # LEGACY, Old psp files use single half_life
    @classmethod
    def from_half_life(cls, half_life):
        return cls(fraction1=1,
                   half_life1=half_life)
        


class Wall(ModelItem):
    label = labels.WALLS

    _attr_dct = {
        "vertices": labels.VERTICES, 
        "shielding": labels.SHIELDING,
    }
    
    _attr_defaults = {
        labels.VERTICES: ZERO_VERTICES,
        labels.SHIELDING: labels.EMPTY_SHIELDING
    }
        
    def set_vertex(self, index, vertex):
        self.logger.debug('Test!!!!!!!')
        # needed by gui
        vertices = self.vertices.copy()
        vertices[index] = vertex
        self.vertices = vertices # generates event for gui if copied first
        

        
class CriticalPoint(ModelItem):
    default_name = 'Critical Point'
    label = labels.CRITICAL_POINTS

    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION, 
        "occupancy_factor": labels.OCCUPANCY_FACTOR,
        'enabled': labels.ENABLED
    }
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.OCCUPANCY_FACTOR: 1,
        labels.ENABLED: True
    }

class SourceNM(ModelItem):
    default_name = 'Source NM'
    label = labels.SOURCES_NM
    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION,  
        "number_of_exams": labels.NUMBER_OF_EXAMS,
        "isotope": labels.ISOTOPE,
        "self_shielding": labels.SELF_SHIELDING,
        "activity": labels.ACTIVITY,
        "duration": labels.DURATION,
        "apply_decay_correction": labels.APPLY_DECAY_CORRECTION,
        'clearance': labels.CLEARANCE,
        'enabled': labels.ENABLED
    }
    
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.NUMBER_OF_EXAMS: 1,
        labels.ISOTOPE: 'F-18',
        labels.SELF_SHIELDING: 'None', # must be str 'None' or 'Body' 
        labels.ACTIVITY: 1,
        labels.DURATION: 1,
        labels.APPLY_DECAY_CORRECTION: True,
        labels.CLEARANCE: labels.EMPTY_CLEARANCE,
        labels.ENABLED: True}
                

    
    @property
    def tiac(self):
        return self.time_integrated_activity_coefficient_mbqh

class SourceXray(ModelItem):
    default_name = 'Source Xray'
    label = labels.SOURCES_XRAY
    _kvp = None
    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION, 
        "number_of_exams": labels.NUMBER_OF_EXAMS,
        "kvp": labels.KVP,
        "dap": labels.DAP,
        "enabled": labels.ENABLED,
    }
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.DAP: 1,
        labels.NUMBER_OF_EXAMS: 1,
        labels.ENABLED: True,
        labels.KVP: CONSTANTS.xray[0].kvp
        }

        

class SourceCT(ModelItem):
    default_name = 'Source CT'
    label = labels.SOURCES_CT
    _attr_dct = {
        "name": labels.NAME,
        "position": labels.POSITION, 
        "number_of_exams": labels.NUMBER_OF_EXAMS,
        "body_part": labels.CT_BODY_PART,
        "kvp": labels.KVP,
        "dlp": labels.DLP,
        'enabled': labels.ENABLED
    }
    
    
    _attr_defaults = {
        labels.NAME: default_name,
        labels.POSITION: ZERO_POSITION,
        labels.DLP: 1,
        labels.NUMBER_OF_EXAMS: 1,
        labels.CT_BODY_PART: 'Body',
        labels.KVP: CONSTANTS.ct[0].kvp,
        labels.ENABLED: True}
        
    
    @property
    def available_kvp(self):
        return list(set([item.kvp for item in CONSTANTS.ct]))
    
   



import numpy as np
from pyrateshield.model_items import Wall
from pyrateshield.floorplan_and_dosemap import Geometry
                                     
from pyrateshield import labels
from pyrateshield.gui import styles
from pyrateshield import Logger, GLOBAL_LOG_LEVEL

LOG_LEVEL = GLOBAL_LOG_LEVEL



class PixelSizeLine(Wall):
    label = labels.PIXEL_SIZE_CM
    
    _attr_dct = {
        "vertices": labels.VERTICES, 
    }
    

class MplItemsPlotter(Logger):
    _model = None
    pixel_size = None
    def __init__(self, axes=None, model=None):
        Logger.__init__(self, log_level=LOG_LEVEL)
        self.plotted_items = []
        self.axes = axes
        self.model = model
        self.pixel_size = self.get_pixel_size()
        self.plotted_items += [self.pixel_size]
        
    def get_pixel_size(self):
        vertices = self.model.floorplan.geometry.vertices
        model_item = PixelSizeLine(vertices=vertices)
        return MPLPixelSize(model_item=model_item,
                                       axes=self.axes)
        
    @property
    def model_items(self):
        return [*self.model.sources_ct, 
                *self.model.sources_nm,
                *self.model.sources_xray, 
                *self.model.critical_points,
                *self.model.walls]
    
    def set_callbacks(self):
        for container in [self.model.sources_ct, 
                          self.model.sources_nm,
                          self.model.sources_xray, 
                          self.model.critical_points,
                          self.model.walls]:
            
            container.connect(self, container.EVENT_ADD_ITEM,
                              self.add_mpl_item)
            
            container.connect(self, container.EVENT_WILL_REMOVE_ITEM,
                              self.remove_model_item)
            
        self.model.item_selector.connect(self, self.model.item_selector.EVENT_SELECT_ITEM, self.select)
    
    def find_item_by_matplotlib(self, matplotlib_item):
        mpl_item = None
        for item in self.plotted_items:
            if item.contains_matplotlib_item(matplotlib_item):
                mpl_item = item
        return mpl_item
    
    def mpl_item(self, model_item):
        mpl_item = None
        for item in self.plotted_items:
            if item.model_item is model_item:
                mpl_item = item                
        return mpl_item
    
               
        
    def disconnect(self):
        if self.model is None: return
        
        for container in (self.model.sources_nm, 
                          self.model.sources_ct,
                          self.model.sources_xray, 
                          self.model.walls,
                          self.model.critical_points,
                          self.model.item_selector):
            container.disconnect(self)
        
    def select(self, model_item):
        self.logger.debug(f'Will Select Model Item of type {type(model_item)}!')
        
        
        for mpl_item in self.plotted_items:
            if mpl_item.model_item is not model_item:
                mpl_item.deselect()
            else:
                self.logger.debug('Item Selected!')
                mpl_item.select()
        
    
    def clear(self):
        for item in self.plotted_items:
            item.remove()

        self.pixel_size = None
        self.plotted_items = []

        
    def refresh(self):
        self.clear()
        for model_item in self.model_items:
            self.add_mpl_item(model_item)
        self.pixel_size = self.get_pixel_size()
        self.plotted_items += [self.pixel_size]
            
        
    def remove_model_item(self, model_item):
        for mpl_item in self.plotted_items:
            if mpl_item.model_item is model_item:
                mpl_item.remove()
                self.plotted_items.remove(mpl_item)
                
    
    def add_mpl_item(self, model_item):
        if isinstance(model_item, Wall):
            mpl_item = MPLWall(model_item=model_item, 
                               shieldings=self.model.shieldings,
                               axes=self.axes)
        else:
            mpl_item = MPLPoint(model_item=model_item, axes=self.axes)
        
        if model_item == self.model.item_selector.select_item:
            mpl_item.select()
        
        self.plotted_items += [mpl_item]
        
    
                         
            
    @property
    def model(self):
        return self._model
    
    @model.setter
    def model(self, model):
        if model is self.model:
            return
        if self._model is not None:
            self.disconnect()
            
        self._model = model
        self.set_callbacks()
        self.refresh()
        self.pixel_size.deselect()
        
    
class MPLPoint(Logger):
    selected = False
    enabled = False
    _pick_enabled = True
    
    def __init__(self, model_item=None, axes=None):
        Logger.__init__(self)
        self.axes = axes
        self.model_item = model_item
        self.plot()
        
        self.model_item.connect(self, self.model_item.EVENT_UPDATE,
                                self.update)
    def get_pick_enabled(self):
        return self._pick_enabled
    
    def enable_pick(self):
        self._pick_enabled = True

    def disable_pick(self):
        self._pick_enabled = False
    
    def remove(self):
        self.model_item.disconnect(self)
        self.mpl_item.remove()
        self.draw()
    
    def plot(self):
        if self.model_item.enabled:
            style = styles.STYLES[self.model_item.label][styles.DEFAULT]
            self.enabled = True
        else:
            style = styles.STYLES[self.model_item.label][styles.DISABLED]
            self.enabled = False
        
        X, Y = self.model_item.position
        
        self.mpl_item = self.axes.plot(X, Y, **style)[0]
        self.draw()
        
    def update(self, event_data):
        model_item, label, old_value, value = event_data
        
        if label == labels.POSITION:
            self.mpl_item.set_xdata(value[0])
            self.mpl_item.set_ydata(value[1])
            
        if label == labels.ENABLED:
            if value:
                self.enable()
            else:
                self.disable()
        self.draw()
                
    def select(self):
        if self.selected:
            return
        
        style = styles.STYLES[self.model_item.label][styles.SELECTED]
        self.mpl_item.update(style)
        self.selected = True
        self.logger.debug(f'Selected Point Item: {self.model_item.name}')
        self.draw()
        
    def deselect(self):
        if not self.selected:
            return
        
        if self.model_item.enabled:
            style = styles.STYLES[self.model_item.label][styles.DEFAULT]
            self.mpl_item.update(style)
        else:
            style = styles.STYLES[self.model_item.label][styles.DISABLED]
            self.mpl_item.update(style)
        self.enabled = self.model_item.enabled
        self.selected = False
        self.draw()
                
    def enable(self):
        if self.enabled:
            return
        
        style = styles.STYLES[self.model_item.label][styles.DEFAULT]
        self.mpl_item.update(style)
        self.enabled = True
        self.draw()
        
    def disable(self):
        if not self.enabled:
            return
        style = styles.STYLES[self.model_item.label][styles.DISABLED]
        self.mpl_item.update(style)
        self.enabled = False
        self.draw()
    
    def contains_matplotlib_item(self, mpl_item):
        return mpl_item == self.mpl_item
        
    
    def draw(self):
        self.axes.figure.canvas.draw_idle()    
        
class MPLLine(Logger):
    selected = False
    _pick_enabled = True
    def __init__(self, model_item=None, axes=None):
        Logger.__init__(self, log_level=LOG_LEVEL)
        self.axes = axes
        
        self.model_item = model_item
        self.plot()
        self.model_item.connect(self, self.model_item.EVENT_UPDATE,
                                self.update)
    def get_pick_enabled(self):
        return self._pick_enabled
    
    def enable_pick(self):
        self._pick_enabled = True

    def disable_pick(self):
        self._pick_enabled = False
        
    def update(self, event_data):
         
         model_item, label, old_value, value = event_data
         self.logger.debug(f'Event: {str(event_data)}')
         if label == labels.VERTICES:
             self.move(value)         
             self.draw()
             
    def remove(self):
        self.model_item.disconnect(self)
        self.mpl_item.remove()
        self.draw()
        
    def plot(self):
        self.plot_line()
        self.plot_markers()
        self.draw()
        
    def plot_line(self, style=None):
        if style is None:
            style = styles.STYLES[self.LINE_LABEL][styles.DEFAULT]
        
        # Plot Line on Axes
 
        X = (self.model_item.vertices[0][0], self.model_item.vertices[1][0])
        Y = (self.model_item.vertices[0][1], self.model_item.vertices[1][1])   
            
        self.mpl_item = self.axes.plot(X, Y, **style)[0]
    
    def plot_markers(self, style=None):
        if style is None:
            style = styles.STYLES[self.MARKER_LABEL][styles.DEFAULT]
            
        X = (self.model_item.vertices[0][0], self.model_item.vertices[1][0])
        Y = (self.model_item.vertices[0][1], self.model_item.vertices[1][1])
        
        marker1 = self.axes.plot(X[0], Y[0], **style)[0]
        marker2 = self.axes.plot(X[1], Y[1], **style)[0]
        self.markers = (marker1, marker2)
        
    def move(self, value):
        self.move_line(value)
        self.move_markers(value)
        self.draw()
        
    def move_line(self, value):
        self.mpl_item.set_xdata((value[0][0], value[1][0]))
        self.mpl_item.set_ydata((value[0][1], value[1][1]))
        
    def move_markers(self, value):
        self.markers[0].set_xdata(value[0][0],)
        self.markers[0].set_ydata(value[0][1],)
        
        self.markers[1].set_xdata(value[1][0],)
        self.markers[1].set_ydata(value[1][1],)
    
    def select(self):
        self.logger.debug('MPLWall Item Selected!')
        
        # if self.selected:
        #     return
        
        style = styles.STYLES[self.LINE_LABEL][styles.SELECTED]
        self.mpl_item.update(style)
         
        style = styles.STYLES[self.MARKER_LABEL][styles.SELECTED]
        self.markers[0].update(style)
        self.markers[1].update(style)
        self.selected = True
        
        
        self.draw()
        
    def deselect(self):
        if not self.selected:
            return
        self.logger.debug('MPLWall Item Selected!')
        style = styles.STYLES[self.LINE_LABEL][styles.DEFAULT]
        self.mpl_item.update(style)
        
        style = styles.STYLES[self.MARKER_LABEL][styles.DEFAULT]
        self.markers[0].update(style)
        self.markers[1].update(style)
        self.selected=False
        self.draw()
        
        
    def contains_matplotlib_item(self, mpl_item):
        return mpl_item is self.mpl_item or mpl_item in self.markers
        
    def draw(self):
        self.axes.figure.canvas.draw_idle()
        
    
    def get_vertex_index_at_position(self, position, tolerance_cm=0):
        # needed by gui
        if tolerance_cm == 0:
            if position not in self.model_item.vertices:
                return None
            else:
                return self.model_item.vertices.index(position)
        else:
            # check if position is close enough to vertices
            v1, v2 = self.model_item.vertices
            d1 = np.linalg.norm(np.asarray(v1) - np.asarray(position))
            d2 = np.linalg.norm(np.asarray(v2) - np.asarray(position))
            
            if d1 > tolerance_cm and d2 > tolerance_cm:
                # vertices to far away from position
                return None
            else:
                # return vertex index with closest distance
                return [d1, d2].index(min([d1, d2]))
            
        

class MPLWall(MPLLine):
    LINE_LABEL = labels.WALLS
    MARKER_LABEL = labels.WALL_MARKER
    
    def __init__(self, model_item=None, shieldings=None, axes=None):
        self.shieldings = shieldings
        super().__init__(model_item=model_item, axes=axes)
        
    def update(self, event_data):
        model_item, label, old_value, value = event_data
        if label in (labels.SHIELDING, labels.COLOR, labels.LINEWIDTH):
       
            fm = {'color': self.shielding.color, 
                 'linewidth': self.shielding.linewidth}
            self.mpl_item.update(fm)
        
            if label == labels.SHIELDING or label == labels.COLOR:
                self.markers[0].update({'color': self.shielding.color})
                self.markers[1].update({'color': self.shielding.color})
            
            self.draw()
            
        super().update(event_data)
    
    @property
    def shielding(self):
        name = self.model_item.shielding
        return self.shieldings.get_item_by_name(name)
        
    def plot_line(self, _=None):
        style = styles.STYLES[labels.WALLS][styles.DEFAULT].copy()
        style['color'] = self.shielding.color
        style['linewidth'] = self.shielding.linewidth
        super().plot_line(style=style)
        
    def plot_markers(self, _=None):
        # Plot Selection Markers, only visible when item is selected
        style = styles.STYLES[labels.WALL_MARKER][styles.DEFAULT].copy()
        style['color'] = self.shielding.color
        super().plot_markers(style=style)
        
   


            
class MPLPixelSize(MPLLine):
    LINE_LABEL = labels.PIXEL_SIZE_CM
    MARKER_LABEL = labels.PIXEL_SIZE_MARKER
    



    
if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from pyrateshield.model import Model
    axes = plt.axes()
    project = '/Users/marcel/git/pyrateshield/example_projects/SmallProject/project.psp'
    model = Model.load_from_project_file(project)
    
    items = MplItemsPlotter(model=model, axes=axes)
    
    
       
                
   
                       
   
            
            
   

    
    
    
       
            
    
    
            
        
        

        




# -*- coding: utf-8 -*-
"""
Isotope calculations for pyshield package

Last Updated 05-02-2016
"""
import numpy as np
import scipy.interpolate as si
from pyrateshield.pyshield import tables

         

DEBUG = False



make_list = lambda item: item if isinstance(item, (list, tuple)) else [item]

    


def attenuation(energy_keV, material, thickness):
    """
    Attenuation for a given energy through a matrial with thickness.
    Args:
        energy_keV: the energy of  the photon in keV
        material: name of the material
        thickness: thickness of the material
    Returns:
        a:  attenation factor (float)
    """
    thickness = np.asarray(thickness)
    
    a = np.exp(-u_linear(energy_keV, material) * thickness)

    return a


def u_linear(energy_keV, material):
    """
    Args:
      energy_keV: the energy of  the photon in keV
      material: name of the material
    Returns:
      Linear attenuation coefficient in [cm^-1]
    Raises:
      NameError if material is not defined in the pyshield recources
    """

    mu_p_i = u_mass(energy_keV, material)

    return mu_p_i * material.density

def u_mass(energy_keV, material):
    try:
        table = tables.ATTENUATION_TABLES[material.attenuation_table]
    except NameError:
        raise NameError(material.attenuation_table + ' not in attenuation table!')

    energies = np.array(table[tables.ENERGY_MeV])

    mu_p = np.array(table[tables.MASS_ATTENUATION])

    interp_fcn = si.interp1d(energies, mu_p)

    mu_p_i = interp_fcn(energy_keV / 1e3)

    return mu_p_i

def interp2d_pairs(*args,**kwargs):
    # https://stackoverflow.com/questions/47087109/evaluate-the-output-from-scipy-2d-interpolation-along-a-curve
    """ Same interface as interp2d but the returned interpolant will evaluate its inputs as pairs of values.
    """
    # Internal function, that evaluates pairs of values, output has the same shape as input
    def interpolant(x,y,f):
        x,y = np.asarray(x), np.asarray(y)
        return (si.dfitpack.bispeu(f.tck[0], f.tck[1], f.tck[2], f.tck[3], f.tck[4], x.ravel(), y.ravel())[0]).reshape(x.shape)
    # Wrapping the scipy interp2 function to call out interpolant instead
    return lambda x,y: interpolant(x,y,si.interp2d(*args,**kwargs))

    # # Create the interpolant (same interface as interp2d)
    # f = interp2d_pairs(X,Y,Z,kind='cubic')
    # # Evaluate the interpolant on each pairs of x and y values
    # z=f(x,y)

# def buildup(energy_keV, material, thickness):
#     """
#     Buildup for a given energy through a matrial with thickness.
#     Args:
#         energy_keV: the energy of  the photon in keV
#         material: name of the material
#         thickness: thickness of the material
#     Returns:
#         b:  buildup factor (float)
#     """
#     # if thickness == 0:
#     #     return 1
    
    
#     if isinstance(thickness, (float, int)) or thickness.ndim == 0:
#         thickness = [float(thickness)]
    
#     thickness = np.asarray(thickness)
    
#     index = thickness > 0

#     return BuildupHelper.calculate(material, energy_keV, thickness)
#     try:
#         table = PHYSICS[BUILDUP][material.buildup_table]
#     except NameError:
#         raise NameError(material.name + ' not in buildup table!')
    
#     n_mfp       = np.asarray(tables.BUILDUP_TABLES[tables.MFP], 'float64')
#     table       = table.drop(tables.MFP, axis=1)
    
#     factors     = np.asarray(table, 'float64')
#     energies    = np.asarray(table.columns, dtype='float64')
    
#     n_mfpi      = number_mean_free_path(energy_keV, 
#                                       material, 
#                                       thickness[index])

#     interp_func2d = interp2d(energies, n_mfp, factors)
#     interp_func1d = lambda ii: interp_func2d(energy_keV/1000, ii)
    
#     n_mfpi    = number_mean_free_path(energy_keV, material.name, 
#                                       thickness[index])#[index])
    
#     buildup_values = np.asarray([float(interp_func1d(ii)) for ii in n_mfpi])
    
#     buildup = np.ones(len(thickness))
#     buildup[index] = buildup_values.flatten()

#     return buildup


def number_mean_free_path(energy_keV, material, thickness):
    """"
    Args:
      energy_keV: the energy of  the photon in keV
      material: name of the material
      thickness: thickness of the material
    Retuns:
      number of mean free paths for a given photon enery, material and
      material thicknesss
    """
    
    # 1 mean free path = 1 / u_lin
    

    return thickness * u_linear(energy_keV, material)




class BuildupHelper:
    _interpolators = None

    
    @staticmethod
    def interpolant(x,y,f):
        # https://stackoverflow.com/questions/47087109/evaluate-the-output-from-scipy-2d-interpolation-along-a-curve
        """ Same interface as interp2d but the returned interpolant will evaluate its inputs as pairs of values.
        """
        x,y = np.asarray(x), np.asarray(y)
        return (si.dfitpack.bispeu(f.tck[0], f.tck[1], f.tck[2], f.tck[3], f.tck[4], x.ravel(), y.ravel())[0]).reshape(x.shape)
        # Wrapping the scipy interp2 function to call out interpolant instead
        #return lambda x,y: interpolant(x,y,si.interp2d(*args,**kwargs))

    @classmethod
    def _get_interpolator(cls, material):
        if cls._interpolators is None:
            cls._interpolators = {}
        
        if material.buildup_table not in cls._interpolators.keys():
            table       = tables.BUILDUP_TABLES[material.buildup_table]
            n_mfp       = np.asarray(table[tables.MFP], 'float64')
            table       = table.drop(tables.MFP, axis=1)
            factors     = np.asarray(table, 'float64')
            energies    = np.asarray(table.columns, dtype='float64')
            interpolator = si.interp2d(energies, n_mfp, factors)
            interp_fcn = lambda x, y: cls.interpolant(x, y, interpolator)
            cls._interpolators[material.buildup_table] = interp_fcn
        return cls._interpolators[material.buildup_table]
    
    @classmethod
    def calculate(cls, material, energy_keV,  thickness):
        # 2DO cutoff at lowest mean free path instead of 0
        interpolator = cls._get_interpolator(material)
        
        if isinstance(thickness, np.ndarray):
            index = thickness > 0
            thickness = thickness[index]
        else:
            index = None
        
        n_mfpi = number_mean_free_path(energy_keV, material, thickness)
        
        if not isinstance(energy_keV, np.ndarray) and index is not None:
            energy_keV = np.ones(len(n_mfpi)) * energy_keV
        
    
        values = interpolator(energy_keV/1000, n_mfpi)
        
        if index is not None:
            buildup = np.ones(len(index))
             
            buildup[index] = values.flatten()
        else:
            buildup = values
        return buildup
 


if __name__ == "__main__":
    isotope = 'F-18'
    
    a = attenuation(511, 'Water', 25)
    b = buildup(511, 'Water', 25)













from PyQt5.QtWidgets import (QPushButton, QLabel,
                             QComboBox, QDoubleSpinBox, QApplication, QTableView)
from PyQt5.QtGui import QIcon, QFont
import qtawesome                             
import pandas as pd

from pyrateshield.constants import CONSTANTS
from pyrateshield.model import DEFAULTS
from pyrateshield.labels import EMPTY_MATERIAL, ENERGY_KEV, ABUNDANCE, PARENT, ISOTOPES
from pyrateshield.gui.item_views import EditViewBase
from pyrateshield.gui.critical_point_controller import PandasModel

INFO_LABEL =('Isotopes can not be added by the user at the moment. '
             'To request a new isotope send an e-mail to the developers '
             ' of pyrateshield.'
             '\n'
             'For some isotopes daughters are included. For daughter nuclei '
             'the abundance is corrected for the branching ratio.'
             '\n'
             'Pyshield neglects gamma rays with energies smaller than 30 keV '
             'because no buildup data is available for these energies.')


class View(EditViewBase):
    explanation = INFO_LABEL
    def create_widgets(self):
        self.isotope_combo = QComboBox()
        self.isotope_label = QLabel()
        self.table_view = QTableView()
        
    def create_layout(self):
        self.setWindowTitle(ISOTOPES)
        icon = qtawesome.icon('mdi.atom')
        self.setWindowIcon(QIcon(icon))
        self.layout.addWidget(self.isotope_combo, 0, 0)
        self.layout.addWidget(self.table_view, 1, 0)

    def set_stretch(self):
        
        
        
        if hasattr(self, 'explanation'):
            label = QLabel(self.explanation)
            label.setWordWrap(True)
            self.layout.addWidget(label)
            
        self.setLayout(self.layout)
        
    @property
    def isotope(self):
        return self.isotope_combo.currentText()

class Controller:
    def __init__(self, view=None):
        if view is None:
            view = View()
        
        self.view = view
        
        
        self.add_callbacks()
        
        
        self.view.isotope_combo.addItems(self.isotopes)
        
    @property
    def model(self):
        return PandasModel(self.data)
        
    @property
    def data(self):
        data = []
        if self.view.isotope != '':
            isotope = CONSTANTS.get_isotope_by_name(self.view.isotope)
            for parent, keV, abundance in isotope.spectrum_with_parent:
                entry = {PARENT: parent, ENERGY_KEV: keV, ABUNDANCE: abundance}
                data += [entry]
        return pd.DataFrame(data)
        
    def add_callbacks(self):
        self.view.isotope_combo.currentIndexChanged.connect(self.update_isotope)
        
    @property
    def isotopes(self):
        return [item.name for item in CONSTANTS.isotopes]
    
    def update_isotope(self):
        self.view.table_view.setModel(self.model)

        
if __name__ == "__main__":
    app = QApplication([])
    
    controller = Controller()
    window = controller.view
    window.show()    
    app.exec_()


import os
import yaml
from pyrateshield import labels
from pyrateshield.modelitem import ModelItem



CONSTANTS_FILE = "constants.yml"
ISOTOPES_FILE = "isotopes.yml"



class Isotope(ModelItem):
    _attr_dct = {
        "name": labels.NAME,
        "half_life": labels.HALF_LIFE,
        "self_shielding_options": labels.SELF_SHIELDING_OPTIONS,
    }
    spectrum = None

class ArcherParamsCT(ModelItem):
    _attr_dct = {
        "kvp": labels.KVP,
        "archer": labels.ARCHER,
        "scatter_fraction_body": labels.CT_SCATTER_FRACTION_BODY,
        "scatter_fraction_head": labels.CT_SCATTER_FRACTION_HEAD,
    }

class ArcherParamsXray(ModelItem):
    _attr_dct = {
        "kvp": labels.KVP,
        "archer": labels.ARCHER,
        "scatter_fraction": labels.XRAY_SCATTER_FRACTION,
    }



class Constants:
    def __init__(self):
        try:
            wdir = os.path.split(__file__)[0] 
        except:
            wdir = os.getcwd()
            
        with open(os.path.join(wdir, CONSTANTS_FILE)) as f:
            constants_yml = yaml.safe_load(f)
            
        with open(os.path.join(wdir, ISOTOPES_FILE)) as f:
            isotope_yml = (yaml.safe_load(f))            
        
        
        
            
        self.isotopes = [Isotope.from_dict(item)\
                         for item in isotope_yml[labels.ISOTOPES]]
        
        decay_chains = isotope_yml[labels.DECAY_CHAINS]
        self.decay_chains = decay_chains
        isotope_spectra = isotope_yml[labels.ISOTOPE_SPECTRA]
        for isotope in self.isotopes:
            spectrum = list(isotope_spectra[isotope.name])
            spectrum_with_parent = [(isotope.name, energy, intensity)\
                                    for energy, intensity in spectrum]
            for daughter, abundance in decay_chains.get(isotope.name, []):
                for energy, intensity in isotope_spectra[daughter]:
                    spectrum.append([energy, intensity*abundance])
                    spectrum_with_parent.append([daughter, energy, intensity*abundance])
            isotope.spectrum = spectrum
            isotope.spectrum_with_parent = spectrum_with_parent
            
        self.ct = [ArcherParamsCT.from_dict(item)\
                   for item in constants_yml[labels.CT_PARAMETERS]]
            
        self.xray = [ArcherParamsXray.from_dict(item)\
                     for item in constants_yml[labels.XRAY_PARAMETERS]]
        
        self.CT_body_part_options = constants_yml[labels.CT_BODY_PART_OPTIONS]
        
        self.self_shielding_pyshield = constants_yml[labels.SELF_SHIELDING_PYSHIELD]
    
        self.self_shielding_options = [labels.SELF_SHIELDING_NONE,
                                       labels.SELF_SHIELDING_BODY, 
                                       labels.SELF_SHIELDING_FACTOR]
        
        self.base_materials = [labels.EMPTY_TABLE] + constants_yml[labels.BASE_MATERIALS]
        self.buildup_materials = [labels.EMPTY_TABLE] + constants_yml[labels.BUILDUP_MATERIALS]
    
    
    def get_isotope_name_list(self):
        return [item.name for item in self.isotopes]
        

    def get_isotope_by_name(self, name):
       
       isotope = [x for x in self.isotopes if x.name == name]
       if len(isotope) == 0:
           raise KeyError(f'No isotope with name {name}')
       return isotope[0]
    
  
    
CONSTANTS = Constants()



# test
import os
from pyrateshield.version import __version__
from pyrateshield.logger import Logger, GLOBAL_LOG_LEVEL
from pyrateshield.observable import Observable

__pkg_root__ = os.path.split(__file__)[0]



from pyrateshield import labels, Observable
from pyrateshield.gui.mpl_view import PointClicker, MovePointByMouse
from pyrateshield.logger import Logger, GLOBAL_LOG_LEVEL
from pyrateshield.floorplan_and_dosemap import Geometry, MeasuredGeometry
from pyrateshield.gui.item_views import safe_set_value_to_widget
from pyrateshield.model_items import Material
from pyrateshield.gui.io import pixelsize_error
import numpy as np

import qtawesome as qta
LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10

class ModelUpdateController(Logger):
    _model = None
    def __init__(self, model=None, view=None, mpl_controller=None,
                 log_level=LOG_LEVEL):        
        super().__init__(log_level=LOG_LEVEL)
        self.view = view
        self.model = model        
        self.mpl_controller = mpl_controller
        self.set_view_callbacks()
         
    def set_view_callbacks(self):
        pass
        
    @property
    def model(self):
        return self._model
    
    @model.setter
    def model(self, model):
        if model is self.model:
            return
        elif self.model is not None:
            self.disconnect_model()
        
        
        self._model = model
        
        self.connect_model()
        
    def model_to_view(self, item=None):
        pass
    
    def connect_model(self):
        pass
    
    def disconnect_model(self):
        self.model.item_selector.disconnect(self)
        
class EditPixelSizeController(ModelUpdateController):
    _geometry = None
    distance_pixels = None
    
    
    def set_view_callbacks(self):
        self.view.measure_button.clicked.connect(self.measure)
        
        self.view.connect(self, self.view.EVENT_VIEW_UPDATE,
                          self.update_pixel_size)
        
        self.view.connect(self, self.view.EVENT_GEOMETRY_RADIO_BUTTON,
                          self.geometry_change)
        
        self.view.confirm_button.clicked.connect(self.confirm)
                                              
        self.refresh()
        
    def refresh(self):
        # HACK, line item may be renewed for new projects etc.
        # call this method before pixel size widget becomes active
        self.model_to_view()
        
        if self not in self.mpl_line_item.model_item.callbacks.keys():
            event = self.mpl_line_item.model_item.EVENT_UPDATE
            callback = self.update_distance_pixels
            self.mpl_line_item.model_item.connect(self, event, callback)

        
        self.logger.debug(self.model.floorplan.geometry.__class__.__name__)
    
    def show_pixelsize_line(self):
        if isinstance(self.model.floorplan.geometry, MeasuredGeometry):
            self.mpl_controller.mpl_items.pixel_size.select()
            self.mpl_controller.mpl_items.pixel_size.enable_pick()
        else:
            self.hide_pixelsize_line()
    
    def hide_pixelsize_line(self):
        self.mpl_controller.mpl_items.pixel_size.deselect()
        self.mpl_controller.mpl_items.pixel_size.disable_pick()
    
                
    def geometry_change(self, _=None):
        if self.view.choose_fixed.isChecked():
            self.hide_pixelsize_line()
        else:
            self.show_pixelsize_line()
    
    def update_pixel_size(self, event_data):
       
        if event_data == labels.REAL_WORLD_DISTANCE_CM:
            self.pixel_size_cm = self.get_measured_pixel_size()
        
    def update_distance_pixels(self, event_data):
        self.logger.debug('Update pixel distance!')
        model_item, label, old_value, value = event_data
        vv = np.asarray(model_item.vertices)
        distance_cm = np.linalg.norm(vv[0] - vv[1])
        distance_pixels = distance_cm / self.model.floorplan.geometry.pixel_size_cm
        
        self.distance_pixels = float(distance_pixels)
        
        
        self.pixel_size_cm = self.get_measured_pixel_size()
        
    @property
    def mpl_line_item(self):
        return self.mpl_controller.mpl_items.pixel_size
        
        
    @property
    def pixel_size_cm(self):
        if self.view.choose_fixed.isChecked():
            return self.view.pixel_size.value()
        else:
            return self.get_measured_pixel_size()
        
    def get_measured_pixel_size(self):
        try:
            return self.distance_cm / self.distance_pixels
        except (ZeroDivisionError, TypeError):
            return 1
            
    @pixel_size_cm.setter
    def pixel_size_cm(self, value):
        if self.view.choose_fixed.isChecked():
            safe_set_value_to_widget(self.view.pixel_size, value)
        else:
            safe_set_value_to_widget(self.view.pixel_size_measured, value)
        
    @property
    def distance_cm(self):
        try:
            return self.view.physical_distance.value()
        except ValueError:
            return None
    
    @distance_cm.setter
    def distance_cm(self, value):
        self.view.physical_distance.setValue(value)
        
    @property
    def distance_pixels(self):
        try:
            return float(self.view.pixel_distance.text())
        except ValueError:
            return 1
        
    @distance_pixels.setter
    def distance_pixels(self, value):
        safe_set_value_to_widget(self.view.pixel_distance, value)
        
        
    def model_to_view(self):
        geometry = self.model.floorplan.geometry
        if isinstance(geometry, MeasuredGeometry):
            #self.view.clear()
            self.view.set_choose_measured()
            self.distance_cm = geometry.distance_cm
            self.distance_pixels = geometry.distance_pixels
            self.pixel_size_cm = geometry.pixel_size_cm
            self.mpl_line_item.model_item.vertices = geometry.vertices            
        else:
            self.view.set_choose_fixed()
            self.pixel_size_cm = geometry.pixel_size_cm
            
            
      
   
    def measure(self, _=None):
        clicker = PointClicker(mpl_controller=self.mpl_controller)
        def draw_geometry_line(position):
            clicker.disconnect(self)
            vertices = [position, position]           
            self.mpl_line_item.model_item.vertices = vertices
            
            
        
            MovePointByMouse(mouse=self.mpl_controller.mouse, 
                             model_item=self.mpl_line_item.model_item,
                             vertex_index=1, hold=False)
            
    
        clicker.connect(self, clicker.EVENT_POINT_CLICKED, draw_geometry_line)
        return
        
    def confirm(self, _=None):
        self.logger.debug('Confirming!')
        geometry = self.model.floorplan.geometry
        if self.view.choose_fixed.isChecked():
            
            if geometry.pixel_size_cm == self.pixel_size_cm:
                return
            else:
                new_geometry = Geometry(pixel_size_cm=self.pixel_size_cm,
                                        origin_cm=geometry.origin_cm)
        else:
            if self.distance_cm <= 0:
                pixelsize_error()
                return
                
            vv = self.mpl_line_item.model_item.vertices
            vvp = [geometry.cm_to_pixels(vi) for vi in vv]
           
            if isinstance(geometry, MeasuredGeometry)\
                and (geometry.vertices_pixels == vvp and\
                    geometry.distance_cm == self.distance_cm):
                    self.logger.debug('Nothing to change!')
                    return
            else:
                self.logger.debug('New Geometry!')
                vvp = [[float(vi[0]), float(vi[1])] for vi in vvp]
                new_geometry = MeasuredGeometry(vertices_pixels=vvp,
                                                distance_cm=self.distance_cm,
                                                origin_cm=geometry.origin_cm)
               
                self.mpl_line_item.model_item.vertices = new_geometry.vertices
        self.model.floorplan.geometry = new_geometry
        
        # by large changes in pixel size matching is necessary
        self.model.match_extent_to_floorplan()
        # explicit refresh is necessary now, fix later with an event
        self.mpl_controller.refresh()
        
        self.show_pixelsize_line()
    
        self.logger.debug(f'New Pixel Vertices {self.mpl_line_item.model_item.vertices}')

    

class EditModelItemControllerBase(ModelUpdateController):
    ITEM_LABEL = 'None'

    def set_view_callbacks(self):
        super().set_view_callbacks()
        self.view.connect(self, self.view.EVENT_VIEW_UPDATE,
                          self.write_to_model)
        
        if len(self.sequence) > 0:
            self.model_to_view(self.sequence[0])
        
    def connect_model(self):
        super().connect_model()
           
        self.sequence.connect(self, 
                              self.sequence.EVENT_UPDATE_ITEM,
                              self.model_update_callback)   
        
        self.sequence.connect(self, 
                              self.sequence.EVENT_REMOVE_ITEM,
                              self.item_removed)
        
        self.sequence.connect(self, 
                              self.sequence.EVENT_ADD_ITEM,
                              self.item_added)
        
        self.model.item_selector.connect(self, 
                                         self.model.item_selector.EVENT_SELECT_ITEM,
                                         self.select_item_callback)
        
        if len(self.sequence) > 0:
            self.model_to_view(self.sequence[0])
            
    def disconnect_model(self):
        super().disconnect_model()
        self.sequence.disconnect(self)
        
        
    def item_added(self, item=None):
        self.model_to_view(item)
        self.refresh()
        
    def item_removed(self, item=None):
         if self.model.item_selector.selected_item in self.sequence:
             self.model_to_view(self.model.item_selector.selected_item)
         elif len(self.sequence) > 0:
             self.model_to_view(self.sequence[0])
         else:
             self.view.clear()
         self.refresh()
        
    
    def select_item_callback(self, item):
        self.logger.debug('Select Item Callback')
        if item in self.sequence:
            self.model_to_view(item)
            
    @property
    def sequence(self):
         return self.model.get_sequence_by_label(self.ITEM_LABEL)
     
     
    def write_to_model(self, event_data=None):
        item = self.get_item_in_view()
        if item is not None:
            item.update_by_dict(event_data)

            
            
    def model_to_view(self, item):
        self.view.disable_connection(self)
        self.logger.debug('Model to View')
        if item not in self.sequence:
            return
        
        if item is None:
            self.view.clear()
        else:
            self.logger.debug(f'Setting item of type {type(item)} to view')
            self.logger.debug(f'{str(item.to_dict())}')
            self.view.from_dict(item.to_dict())
        
        self.view.enable_connection(self)
        self.refresh()

        
    def refresh(self):
        # called by controller when view is selected by user
        self.logger.debug(f'Setting view status for n={len(self.sequence)}')
        if len(self.sequence) == 0:
            self.view.set_enabled(False)
        else:
            self.view.set_enabled(True)
        
        pass
    
    def new(self, **kwargs):
        self.mpl_controller.view.toolbar.deselect()
        self.sequence.add_new_item(**kwargs)

        
        
    def model_update_callback(self, event_data=None):
        self.logger.debug('model_update_callback event_data {str(event_data)}')
        #self.view.disable_connection(self)
        item, label, old_value, value = event_data
        
        if item is not self.model.item_selector.selected_item:
            return

        
        self.view.from_dict({label: value})
        
        
        
        self.logger.debug('model_update_callback finished')
        

class EditNamedModelItemController(EditModelItemControllerBase):  
 
    def set_view_callbacks(self):
        super().set_view_callbacks()            
        self.view.connect(self, self.view.EVENT_LIST_SELECTION, 
                          self.list_selection_callback)
    
    def model_update_callback(self, event_data=None):
        item, label, old_value, value = event_data
        if label == labels.NAME:
            self.update_list()
        
        super().model_update_callback(event_data=event_data)
        
    def connect_model(self):
        super().connect_model()
        self.update_list()
    
    def item_added(self, item=None):
        self.update_list()
        super().item_added(item=item)
       
    def item_removed(self, item=None):
        self.update_list()
        super().item_removed(item=item)
        
    def update_list(self):
        self.view.disable_connection(self)
        current_index = self.view.list.currentIndex()
        self.view.list.clear()
        
        for item in self.sequence:
            self.add_list_item(item)
        
        if self.view.list.count() >= current_index:
            self.view.list.setCurrentIndex(current_index)
            
        self.logger.debug(f'New selected list item: {self.view.list.currentText()}')
        self.view.enable_connection(self)
        
    def add_list_item(self, model_item):
        self.view.list.addItem(model_item.name)
        
    def add_model_item(self, _=None):
        super().add_model_item()
        self.update_list()
        
    def remove_model_item(self, _=None):
        super().remove_model_item()
        self.update_list()
        if len(self.sequence) == 0:
            self.view.set_enabled(False)
        else:
            self.view.set_enabled(True)
        
    def get_item_in_view(self):
        if len(self.sequence) == 0:
            return None
        else:
            name = self.view.list.currentText()
            item = self.get_item_by_name(name)
        return item
    
    @property
    def item_names(self):
        return [item.name for item in self.sequence] 
    
    
    def get_item_by_name(self, name):
        if name not in self.item_names:
            return None
        
        index = self.item_names.index(name) 
    
        return self.sequence[index]
            
    def model_to_view(self, item=None):
        self.view.disable_connection(self)
        if item not in self.sequence:
            return
        #self.view.disable_connection(self)
        self.view.list.setCurrentIndex(self.sequence.index(item))
        super().model_to_view(item)
        self.view.enable_connection(self)
        
    def list_selection_callback(self, index=None):
        self.model.item_selector.select_item(self.sequence[index])
        
    @property
    def list_items(self):
        return [self.view.list.itemText(i)\
                for i in range(self.view.list.count())]


class EditSourceCTController(EditNamedModelItemController):
    ITEM_LABEL = labels.SOURCES_CT
    
class EditSourceXrayController(EditNamedModelItemController):
    ITEM_LABEL = labels.SOURCES_XRAY

class EditCriticalPointsController(EditNamedModelItemController):
    ITEM_LABEL = labels.CRITICAL_POINTS
    
class EditSourcesNMController(EditNamedModelItemController):
    ITEM_LABEL = labels.SOURCES_NM
    
    def connect_model(self):
        super().connect_model()
        
        self.model.clearances.connect(self, 
                                      self.model.clearances.EVENT_UPDATE_ITEM,
                                      self.update_clearance_name)
        
        self.model.clearances.connect(self, 
                                      self.model.clearances.EVENT_ADD_ITEM,
                                      self.update_clearance_list)
        
        self.model.clearances.connect(self, 
                                      self.model.clearances.EVENT_REMOVE_ITEM,
                                      self.update_clearance_list)
        
        self.update_clearance_list()
        

    def disconnect_model(self):
        super().disconnect_model()
        self.model.clearances.disconnect(self)
        
    def update_clearance_name(self, event_data=None):
        model_item, label, old_value, value = event_data
        if label == labels.NAME:
            self.update_clearance_list()
    
    def update_clearance_list(self, event_data=None):
        self.logger.debug('Updating clearance list')
        
        item = self.get_item_in_view()
        
        if item is None:
            current_clearance = None
        else:
            current_clearance = item.clearance
        
        self.view.clearance_list.clear()
        
        for item in self.model.clearances:
            self.view.clearance_list.addItem(item.name)

        if current_clearance is not None:
            safe_set_value_to_widget(self.view.clearance_list, 
                                     current_clearance)
            
                  
    
class EditClearanceModel(EditNamedModelItemController):
    ITEM_LABEL = labels.CLEARANCE
    
    def set_view_callbacks(self):
        super().set_view_callbacks()
        self.view.new_button.clicked.connect(self.new)
        self.view.delete_button.clicked.connect(self.delete)
    
    def refresh(self):
        if len(self.sequence) <= 1 or self.view.list.currentIndex() == 0:
            self.view.set_enabled(False)
        else:
            self.view.set_enabled(True)
        
        self.toggle_delete()
        self.view.toggle_split_time()
        
    def delete(self, _=None):
        name = self.view.to_dict()[labels.NAME]
        index = self.item_names.index(name)
        self.logger.debug(f'Removing item: {index}')
        self.sequence.pop(index)
        self.refresh()
    
    @property
    def used_clearance_models(self):
        names = [labels.EMPTY_CLEARANCE]
        names += [source.clearance\
                       for source in self.model.sources_nm]
        
        return list(set(names))
    
    
    def toggle_delete(self, _=None):
        item = self.get_item_in_view()
        
        if item is None or item.name in self.used_clearance_models:
            self.view.delete_button.setEnabled(False)
        else:
           
            self.view.delete_button.setEnabled(True) 
            
    def select_item_callback(self, item):
        super().select_item_callback(item)
        
        if item in self.model.sources_nm:            
            clearance = self.model.clearances.get_item_by_name(item.clearance)
            self.logger.debug(f'!!! Type {type(self.model.item_selector.selected_item)}')
            self.model_to_view(clearance)
            self.logger.debug(f'!!! Type {type(self.model.item_selector.selected_item)}')

class EditMaterialsController(EditNamedModelItemController):
    ITEM_LABEL = labels.MATERIALS
    
    def set_view_callbacks(self):
        super().set_view_callbacks()
        self.view.new_button.clicked.connect(self.new)
        self.view.delete_button.clicked.connect(self.delete)
        
    def delete(self, _=None):
        item = self.sequence.get_item_by_name(self.view.to_dict()[labels.NAME])
        self.sequence.remove(item)
        self.refresh()
        
    def refresh(self):
        if len(self.sequence) <= 1 or self.view.list.currentIndex() == 0:
            self.view.set_enabled(False)
        else:
            self.view.set_enabled(True)
        
        self.toggle_delete()
    
    def connect_model(self):
        super().connect_model()
        
        
        self.model.materials.connect(self,
                                     self.model.materials.EVENT_ADD_ITEM,
                                     self.toggle_delete)
        
        self.model.materials.connect(self,
                                     self.model.materials.EVENT_REMOVE_ITEM,
                                     self.toggle_delete)
        
    
    
    def toggle_delete(self, _=None):
        item = self.get_item_in_view()
        
        if item is None or item.name in self.model.shieldings.used_materials:
            self.view.delete_button.setEnabled(False)
        else:
           
            self.view.delete_button.setEnabled(True)     
            
    
   
        
        

            
class EditShieldingController(EditNamedModelItemController):
    ITEM_LABEL = labels.SHIELDINGS
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_materials()
    
    def set_view_callbacks(self):
        super().set_view_callbacks()
        self.view.new_button.clicked.connect(self.new)
        self.view.delete_button.clicked.connect(self.delete)


    def refresh(self):
        super().refresh()
        if len(self.sequence) <= 1 or self.view.list.currentIndex() == 0:
            self.view.set_enabled(False)
        else:
            self.view.set_enabled(True)
        
        self.toggle_delete()
            
            
    def connect_model(self):
        super().connect_model()
        
        
        self.model.walls.connect(self,
                                  self.model.walls.EVENT_ADD_ITEM,
                                  self.toggle_delete)
        
        self.model.walls.connect(self,
                                  self.model.walls.EVENT_REMOVE_ITEM,
                                  self.toggle_delete)
        
        self.model.materials.connect(self,
                                     self.model.materials.EVENT_ADD_ITEM,
                                     lambda _: self.update_materials())
        
        self.model.materials.connect(self,
                                     self.model.materials.EVENT_REMOVE_ITEM,
                                     lambda _: self.update_materials())
        
        self.model.materials.connect(self,
                                     self.model.materials.EVENT_UPDATE_ITEM,
                                     self.update_materials)
        
    def update_materials(self, event_data=None):
        
        
        self.view.disable_connection(self)
        
        
        self.view.material1_list.clear()
        self.view.material1_list.addItems(self.model.materials.names)
        
        self.view.material2_list.clear()
        self.view.material2_list.addItems(self.model.materials.names)
            
        self.view.enable_connection()
            
    def disconnect_model(self):
        super().disconnect_model()
        self.model.walls.disconnect(self)
        
    
    def model_update_callback(self, event_data):
        super().model_update_callback(event_data)
        item, label, old_value, value = event_data
        if label == labels.COLOR or label == labels.NAME:
            # update color icons in list
            self.update_list()
        
    def delete(self, _=None):
        item = self.sequence.get_item_by_name(self.view.to_dict()[labels.NAME])
        self.sequence.remove(item)
        self.refresh()
      
    @property
    def used_shieldings(self):
        shieldings = [labels.EMPTY_SHIELDING]
        shieldings += [wall.shielding for wall in self.model.walls]
        return list(set(shieldings))
    
    
        
    def toggle_delete(self, _=None):
        item = self.get_item_in_view()
        
        if item is None or item.name in self.used_shieldings:
            self.view.delete_button.setEnabled(False)
        else:
           
            self.view.delete_button.setEnabled(True)     
    
    def select_item_callback(self, item):
        super().select_item_callback(item)
        
        if item in self.model.walls:            
            shielding = self.model.shieldings.get_item_by_name(item.shielding)
            self.logger.debug(f'!!! Type {type(self.model.item_selector.selected_item)}')
            self.model_to_view(shielding)
            self.logger.debug(f'!!! Type {type(self.model.item_selector.selected_item)}')

    
    def add_list_item(self, model_item):
        icon = qta.icon('fa5s.circle', color=model_item.color)
        self.view.list.addItem(icon, model_item.name)

class EditWallsController(EditModelItemControllerBase):
    ITEM_LABEL = labels.WALLS
   

    def set_view_callbacks(self):
        super().set_view_callbacks()            
        self.view.connect(self, self.view.EVENT_SCROLL, self.update_by_scroll)
        
        
    def get_item_in_view(self):
        if len(self.sequence) == 0:
            return None
        else:
            index = self.view.scroll_widget.value()
            if index < (len(self.sequence)):
                return self.sequence[index]
            else:
                raise IndexError()
    
    def item_added(self, item=None):
        index = self.sequence.index(item)
        self.logger.debug(f'New Wall Added {index}!')
        self.update_scroll_length()
        super().item_added(item)
        
    def item_removed(self, item=None):
        self.update_scroll_length()
        super().item_removed(item=item)
        
    def connect_model(self):

        super().connect_model()
        
        self.update_scroll_length()
        self.update_shielding_list()
        
        self.model.shieldings.connect(self, 
                                      self.model.shieldings.EVENT_UPDATE_ITEM,
                                      self.update_shielding)
        
        
        self.model.shieldings.connect(self, 
                                      self.model.shieldings.EVENT_ADD_ITEM,
                                      self.update_shielding_list)
        
        self.model.shieldings.connect(self, 
                                      self.model.shieldings.EVENT_REMOVE_ITEM,
                                      self.update_shielding_list)
    def disconnect_model(self):
        super().disconnect_model()
        self.model.shieldings.disconnect(self)
      
        
    def update_shielding(self, event_data):
        item, label, old_value, value = event_data
        if label == labels.COLOR or label == labels.NAME:
            self.update_shielding_list()
        

        
    def update_by_scroll(self, index):
        if index < len(self.sequence):
            self.model.item_selector.select_item(self.sequence[index])
    
  
    def update_scroll_length(self):
        
        self.view.scroll_widget.setMaximum(max(0, len(self.sequence)-1))
    
    def model_to_view(self, model=None):
        self.logger.debug('Will Add Wall!')
        self.view.disable_connection(self)
        super().model_to_view(model)
        if model is not None:
            index = self.sequence.index(model)
            safe_set_value_to_widget(self.view.scroll_widget, index)
            index_str = f'Wall index: {str(index)}'
            safe_set_value_to_widget(self.view.index_label, index_str)
        self.view.enable_connection(self)
   
        
    def update_shielding_list(self, event_data=None):
        self.view.disable_connection(self)
        item = self.get_item_in_view()
        
        if item is None:
            self.logger.debug('No item selected!')
            current_shielding = None
        else:
            current_shielding = item.shielding
        
        self.view.shielding_list.clear()
        
        for item in self.model.shieldings:
            icon = qta.icon('fa5s.circle', color=item.color)
            self.view.shielding_list.addItem(icon, item.name)

        self.view.enable_connection(self)        
        
        if current_shielding is not None:
            safe_set_value_to_widget(self.view.shielding_list, current_shielding)

        
          
        
   
      
        
        


from PyQt5.QtWidgets import QMenu, QAction
from PyQt5.QtGui import QCursor
from pyrateshield import labels, Logger, GLOBAL_LOG_LEVEL
from functools import partial
import pyperclip
import yaml
from copy import copy

from pyrateshield.model_items import Wall
from pyrateshield.gui.mpl_items_controller import PixelSizeLine
import qtawesome as qta
SNAP_RADIUS_POINTS = 100

LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10

class ContextMenuController(Logger):
    to_copy = None
    
    def __init__(self, model=None, view=None, mpl_controller=None,
                 main_controller=None):
        Logger.__init__(self, log_level=LOG_LEVEL)
        
        self.view = view
        self.model = model
        self.main_controller = main_controller
        self.mpl_controller = mpl_controller
        
        self.mpl_controller.connect(self, mpl_controller.EVENT_RIGHT_CLICK,
                                    self.right_click)
    
    @property
    def snap_radius_cm(self):
        T =  self.mpl_controller.axes.transData.inverted().transform
        snap_radius_cm = (T([SNAP_RADIUS_POINTS, 0]) - T([0, 0]))[0]
        return snap_radius_cm
    
    def right_click(self, event_data):
        self.logger.debug('Right click event triggered')
        model_item, vertex_index, position = event_data
        self.logger.debug(f'At position {position}')
        
        if model_item is None:
            self.logger.debug('Show canvas context menu')
            self.show_canvas_context_menu(position)
        else:
            self.logger.debug(f'Show context menu for {type(model_item)}')
            self.show_pick_context_menu(model_item, vertex_index, position)
    
    def show_canvas_context_menu(self, position=None):
        context_menu = QMenu(self.view)
       
        for label in [labels.SOURCES_CT, labels.SOURCES_XRAY,
                      labels.SOURCES_NM, labels.CRITICAL_POINTS]:
            
            title = 'Add ' + label + ' item'
            action = context_menu.addAction(title)
            callback = partial(self.add_item, label, position)
            
            action.triggered.connect(callback)
            
        action = context_menu.addAction('Start wall here')
        callback = partial(self.mpl_controller.add_wall_by_mouse, position)
        action.triggered.connect(callback)
        
            
        action = context_menu.addAction('Paste item')
        callback = partial(self.paste_item, position)
        action.triggered.connect(callback)
        
        
        if self.to_copy is None or isinstance(self.to_copy, Wall):
            # pasting walls doesnt make sense
            action.setEnabled(False)
        
        context_menu.exec_(QCursor.pos())
            
    def add_item(self, label, position):
        self.mpl_controller.view.toolbar.button_checked(label, True)
        self.mpl_controller.add_model_item(label, position)                
  
                    
    def get_snap_action(self, model_item, vertex_index, position):
        
        self.logger.debug(f'Get snap action At position {position}')
        
        get_closest = self.mpl_controller.mpl_picker.get_closest_vertex
        
        walls = [wall for wall in self.model.walls if wall is not model_item]
        wall, closest_vertex_index, distance = get_closest(position, walls=walls)
        
        if distance is not None and distance <= self.snap_radius_cm:
                
            closest_vertex = wall.vertices[closest_vertex_index]
        
            disp_vv = str([round(vi) for vi in closest_vertex])
        
            shielding = self.model.shieldings.get_item_by_name(wall.shielding)
            
            action = "Snap to: " + shielding.name + ' at ' + disp_vv
            snap_action = QAction(action)
        
            callback = partial(self.snap_wall, model_item, 
                               vertex_index, closest_vertex)
                               
        
            snap_action.triggered.connect(callback)
        
            icon = qta.icon('fa5s.circle', color=shielding.color)
            snap_action.setIcon(icon)

        else:
            snap_action = QAction('Snap to: ')
            snap_action.setEnabled(False)
        return snap_action
        
    
    def show_pick_context_menu(self, model_item, vertex_index, position):
        context_menu = QMenu(self.view)
        self.logger.debug(f'Generate contextmenu At position {position}')
        if isinstance(model_item, PixelSizeLine):
            return
        
        if not isinstance(model_item, Wall):
            enabled_action = context_menu.addAction("Enabled")
            enabled_action.setCheckable(True)            
            enabled_action.setChecked(model_item.enabled)            
            enabled_action.triggered.connect(model_item.enable)
            
            cut_action = context_menu.addAction("Cut")
            callback = lambda: self.cut_item(model_item)
            cut_action.triggered.connect(callback)
                
        copy_action = context_menu.addAction("Copy")
        callback = lambda: self.copy_item(model_item)
        copy_action.triggered.connect(callback)
            
                    
        delete_action = context_menu.addAction("Delete")
        callback = model_item.delete
        delete_action.triggered.connect(callback)
                
        if isinstance(model_item, Wall) and vertex_index is not None:
            wall_draw_action = context_menu.addAction("Continue wall here")
            callback = partial(self.mpl_controller.add_wall_by_mouse, model_item.vertices[vertex_index])
            wall_draw_action.triggered.connect(callback)
            
            snap_action = self.get_snap_action(model_item, vertex_index, 
                                               position)
                                               
            context_menu.addAction(snap_action)

            
        context_menu.exec_(QCursor.pos())
    
        
    def snap_wall(self, wall, vertex_index, closest_vertex):
        self.logger.debug(f'Snapping wall {wall.vertices}, index {vertex_index} to {closest_vertex}')
        wall.set_vertex(vertex_index, closest_vertex)
     
        

        
        
    def copy_item(self, item):
        self.to_copy = copy(item)
        
        # copies yaml definition of object to clipboard :)
        yaml_text = yaml.dump(self.to_copy.to_dict(), default_flow_style=None)
        pyperclip.copy(yaml_text)
        
        
    def paste_item(self, pos):
        self.to_copy.position = pos
        self.model.add_item(self.to_copy)
        self.model.item_selector.select_item(self.to_copy)
        self.to_copy = None
        
    
    def cut_item(self, item):
        self.copy_item(item)
        sequence = self.model.sequence_for_item(item)
        sequence.remove(item)
        
        
    def delete_item(self, item):
        self.model.delete_item(item)


import numpy as np
from pyrateshield.gui.main_view import MainView, TOOLBOX_LAYOUT
from pyrateshield.model import Model
from PyQt5.QtWidgets import QApplication
from pyrateshield.gui.mpl_controller import MplController
from pyrateshield.dosemapper import Dosemapper
from pyrateshield.model_items import Wall
from pyrateshield.gui.context_menu import ContextMenuController
from pyrateshield.gui.critical_point_controller import CriticalPointReportController
from pyrateshield.floorplan_and_dosemap import Floorplan, Geometry
from pyrateshield.gui import io, calculator, dosemap_style, isotopes

from pyrateshield.gui.item_controllers import (EditSourceXrayController, 
                                               EditSourceCTController, 
                                               EditWallsController, 
                                               EditCriticalPointsController, 
                                               EditSourcesNMController, 
                                               EditShieldingController,
                                               EditClearanceModel,
                                               EditPixelSizeController,
                                               EditMaterialsController)
                                               
                                               
from pyrateshield import labels


    
class MainController():
    _mpl_controller = None
    _critical_points_controller = None
    _calculator = None
    _dosemap_style = None
    
    def __init__(self, dosemapper=None, model=None, view=None):
        
        if dosemapper is None:
            dosemapper = Dosemapper(multi_cpu=False)
        
        if model is None:
            model = Model()
    
        if view is None:
            view = MainView()
        
        self.dosemapper = dosemapper
        self.view = view
        
        if model is None:
            model = self.get_empty_model()
        
        self._model = model
        self.connect_model()
        
        self.controllers = self.create_controllers()
        
        self.set_view_callbacks()
        
        self.context_menu_controller = ContextMenuController(view=self.view,
                             mpl_controller=self.mpl_controller,
                             model=model,
                             main_controller=self)
        
        self.controllers[labels.SHIELDINGS].view.setEnabled(True)
        
        
    @property
    def model(self):
        return self._model
    
    @model.setter
    def model(self, model):
        if model is self.model:
            return
        elif self.model is not None:
            self.disconnect_model()
            
        self._model = model
        self.load_model()
        self.connect_model()
        
    def disconnect_model(self):
        self._mpl_controller = None
        self.model.item_selector.disconnect(self)
    
    def load_model(self):
        for key, controller in self.controllers.items():
            controller.model = self.model

        self.mpl_controller.model = self.model
        self.context_menu_controller.model = self.model
        self.crit_point_controller.model = self.model
        
        
    def connect_model(self):
        self.model.item_selector.connect(self, 
                                         self.model.item_selector.EVENT_SELECT_ITEM,
                                         self.select_item_in_view)
        
    def select_item_in_view(self, model_item):
        if model_item is not None:
            self.view.set_focus(model_item.label)
        
    def set_view_callbacks(self):
        for label, tab in self.view.toolbox_tabs.items():            
            if len(TOOLBOX_LAYOUT[label]) > 1:                
                tab.currentChanged.connect(self.tab_selected)
                
        self.view.toolbox.currentChanged.connect(self.toolbox_selected)

        # {develop}
        mouse = self.mpl_controller.mouse
        mouse.connect(self, mouse.MOVE_EVENT, self.view.set_status_text)
        
        toolbar = self.view.views[labels.CANVAS].toolbar
        toolbar.connect(self, toolbar.EVENT_TOOLBUTTON_CLICK, 
                        self.toolbar_callback)    
        

      
    @property
    def controller_types(self):
        return {
                labels.SOURCES_CT:        EditSourceCTController,
                labels.SOURCES_XRAY:      EditSourceXrayController,
                labels.SOURCES_NM:        EditSourcesNMController,
                labels.CLEARANCE:         EditClearanceModel,
                labels.WALLS:             EditWallsController,
                labels.SHIELDINGS:        EditShieldingController,
                labels.CRITICAL_POINTS:   EditCriticalPointsController,
                labels.PIXEL_SIZE_CM:     EditPixelSizeController,
                labels.MATERIALS:         EditMaterialsController}       

        
    def create_controllers(self):
        self.mpl_controller = MplController(dosemapper=self.dosemapper,
                                            view=self.view.views[labels.CANVAS],
                                            model=self.model)
        
        self.crit_point_controller = CriticalPointReportController(
            view=self.view.views[labels.CRITICAL_POINT_REPORT_VIEW], model=self.model, 
            controller=self, dosemapper=self.dosemapper)
        
    
        controllers = {}
        for key in (labels.SOURCES_NM, labels.SOURCES_CT, 
                   labels.CLEARANCE, labels.SOURCES_XRAY, labels.WALLS, 
                   labels.CRITICAL_POINTS, labels.PIXEL_SIZE_CM, 
                   labels.SHIELDINGS, labels.MATERIALS):

            contr_type = self.controller_types[key]
            view = self.view.views[key]
            controllers[key] = contr_type(model=self.model, 
                                          view = view,
                                          mpl_controller=self.mpl_controller)
        return controllers

    
        
    def toolbox_selected(self, _=None):
        self.tab_selected()
        
        
    def tab_selected(self, _=None):
        label = self.view.get_active_tool_panel_name()
        selected_item = self.model.item_selector.selected_item
        
        if label in [labels.SHIELDINGS] and isinstance(selected_item, Wall):
            wall = self.controllers[labels.WALLS].get_item_in_view()
            if wall is not None:
                shielding = self.model.shieldings.get_shielding_by_name(
                    wall.shielding)
                self.controllers[label].model_to_view(shielding)
        
        if label in [labels.CLEARANCE]:
            source_nm = self.controllers[labels.SOURCES_NM].get_item_in_view()
            if source_nm is not None:
                clearance = self.model.clearances.get_item_by_name(
                    source_nm.clearance)
                self.controllers[label].model_to_view(clearance)
        
        if label != labels.PIXEL_SIZE_CM:
            model_item = self.controllers[label].get_item_in_view()

            self.controllers[labels.PIXEL_SIZE_CM].hide_pixelsize_line()
            self.model.item_selector.select_item(model_item)
            

        else:
            self.controllers[labels.PIXEL_SIZE_CM].show_pixelsize_line()
            
            
        self.controllers[label].refresh()
        



    def toolbar_callback(self, event_data):
        toolname, checked = event_data
    
        if toolname == 'side_panel':
            self.view.toolbox.setVisible(checked)
        elif toolname == 'save_project_as':
            self.save_as()
        elif toolname == 'load_project':
            self.load()
        elif toolname == 'save_project':
            self.save()
        elif toolname == 'load_image':
            self.load_image()
        elif toolname == 'new':
            self.new()
        elif toolname == 'calculator':
            self.show_calculator()
        elif toolname == 'dosemap_style':
            self.show_dosemap_style()
        elif toolname == labels.ISOTOPES:
            self.show_isotopes()
           
          
            
    def load_image(self):
       image = io.safe_load_image()
       if image is not None:
           # Load floorplan and reset pixel size (geometry)
           self.model.floorplan.image = image
                              
           
           # self.controllers[labels.PIXEL_SIZE_CM].refresh()
           self.view.set_focus(labels.PIXEL_SIZE_CM)
           
    def new(self):        
        confirm = io.confirm_new_project()
        if confirm:
            self.model = self.get_empty_model()
        self.view.set_focus(labels.SOURCES_NM)
        
            
    def save(self):
        io.safe_write_model(model=self.model, 
                                        filename=self.model.filename)
            
    def save_as(self):
        io.safe_write_model(model=self.model)
               
    def load(self):
        model = io.safe_load_model(self.model.filename)
        
        if model is not None:        
 
            self.model = model
   
            
    def show_calculator(self):
        self._calculator = calculator.Controller()
        self._calculator.view.show()
    
    def show_dosemap_style(self):
        self._dosemap_style = dosemap_style.Controller(self.model)
        self._dosemap_style.view.show() 
        
    def show_isotopes(self):
        self._isotopes = isotopes.Controller()
        self._isotopes.view.show()
        
        
    def get_empty_model(self):
        default_empty_size_y = 1E2 # empty canvas number of pixels in y
        default_empty_size_cm = (1E4, 1E4) # default new size in cm
        
        shape_cm = (default_empty_size_cm)
        
        
        pixel_size = shape_cm[1] / default_empty_size_y
        
              
        # rounds pixels in x down, should not matter for large arrays
        shape_pixels = (int(shape_cm[0] / pixel_size), 
                        int(shape_cm[1] / pixel_size), 3)

        empty_canvas = np.ones(shape_pixels)
        
        geometry = Geometry(pixel_size_cm=pixel_size,
                            locked=True)
    
        floorplan_model = Floorplan(image=empty_canvas, geometry=geometry)
                                    
        
        return Model(floorplan=floorplan_model)
        

       
def main(model=None, controller=None, dm=None):
        app = QApplication([])
        
        controller = MainController(model=model, dosemapper=dm)
        window = controller.view
        window.show()    
        app.exec_()
    
if __name__ == "__main__":  
    from pyrateshield.dosemapper import Dosemapper

    
    #project = '../../example_projects/SmallProject/project2.zip'
    
    project = '../../example_projects/LargeProject/project.zip'
    
    model = Model.load_from_project_file(project)
    
    app = QApplication([])
    
    controller = MainController(model=None, dosemapper=None)
    window = controller.view
    window.show()    
    app.exec_()
        




import os
import zipfile
import io
import yaml
import pickle
import copy
import imageio

from pyrateshield.constants import CONSTANTS
from pyrateshield import model_items
from pyrateshield.model_sequences import (ClearanceItemSequence, 
                                          ModelItemSequence, 
                                          NamedModelItemSequence, 
                                          ShieldingItemSequence,
                                          MaterialItemSequence)

from pyrateshield.floorplan_and_dosemap import Dosemap, Floorplan, DosemapStyle
from pyrateshield import labels, Observable, GLOBAL_LOG_LEVEL



LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10

class Defaults:
    def __init__(self):
        try:
            wdir = os.path.split(__file__)[0] 
        except:
            wdir = os.getcwd()
            
        self._defaults = yaml.safe_load(open(os.path.join(wdir, 'defaults.yml')))
        
        clearances = self._defaults[labels.CLEARANCE]
        clearances = [model_items.Clearance.from_dict(item) for item in clearances]
        self.clearances = clearances
        
        materials = self._defaults[labels.MATERIALS]
        materials = [model_items.Material.from_dict(item) for item in materials]
        self.materials = materials
        
        shieldings = self._defaults[labels.SHIELDINGS]
        shieldings = [model_items.Shielding.from_dict(item) for item in shieldings]
        self.shieldings = shieldings
        

DEFAULTS = Defaults()


class Model(Observable):
    _floorplan = None
    _selected_item = None
    filename = None
    
    _sources_nm = None
    _sources_ct = None
    _sources_xray = None
    _walls = None
    _clearances = None
    _shieldings = None
    _critical_points = None
    _floorplan = None
    _dosemap = None
    _dosemap_style = None
    _materials = None
    
    EVENT_UPDATE_FLOORPLAN = 'event_update_floorplan'    
    
    def __init__(self, floorplan=None, dosemap=None, dosemap_style=None, 
                 shieldings=None, walls=None, critical_points=None, 
                 sources_ct=None, sources_nm=None, sources_xray=None,
                 clearances=None, materials=None):  
        

        Observable.__init__(self, log_level=LOG_LEVEL)
        # create empty model possibility for gui
        
        
        self.constants = CONSTANTS
        
        self._floorplan = floorplan
        self._dosemap = dosemap
        self._dosemap_style = dosemap_style
        
        
            
        if walls is not None:
            self.walls.extend(walls)
            
        if critical_points is not None:
            self.critical_points.extend(critical_points)
        
        if sources_ct is not None:
            self.sources_ct.extend(sources_ct)
            
        if sources_xray is not None:
            self.sources_xray.extend(sources_xray)

        if sources_nm is not None:
            self.sources_nm.extend(sources_nm)
            
        if clearances is not None:
            self.clearances.extend(clearances)
        else:
            self.clearances.extend(DEFAULTS.clearances)
            
        if materials is not None:
            self.materials.extend(materials)
        else:
            self.materials.extend(DEFAULTS.materials)
        
        if shieldings is not None:
            self.shieldings.extend(shieldings)
        else:
            self.shieldings.extend(DEFAULTS.shieldings)
            
        # force origin at (0, 0), shift older pyrateshield models
        self.zero_origin()
        self.match_extent_to_floorplan()
        self.set_callbacks()
        
        self.item_selector = ModelItemSelector(self)
        

            
            
        
        
    @property
    def floorplan(self):
        if self._floorplan is None:
            self._floorplan = Floorplan()
        return self._floorplan
    
    @floorplan.setter
    def floorplan(self, floorplan):
        self._floorplan = floorplan
    
    @property
    def dosemap(self):
        if self._dosemap is None:
            self._dosemap = Dosemap()
        return self._dosemap
            
    @property
    def dosemap_style(self):
        if self._dosemap_style is None:
            self._dosemap_style = DosemapStyle()
        return self._dosemap_style
        
    @property
    def walls(self):
        if self._walls is None:
            self._walls = ModelItemSequence(item_class=model_items.Wall)
        return self._walls
    
    @property
    def shieldings(self):
        if self._shieldings is None:
            self._shieldings = ShieldingItemSequence(item_class=model_items.Shielding)
        return self._shieldings
    
    @property
    def critical_points(self):
        if self._critical_points is None:
            self._critical_points = NamedModelItemSequence(item_class=model_items.CriticalPoint)
        return self._critical_points
    
    @property
    def sources_nm(self):
        if self._sources_nm is None:
            self._sources_nm = NamedModelItemSequence(item_class=model_items.SourceNM)
        return self._sources_nm
    
    @property
    def clearances(self):
        if self._clearances is None:
            self._clearances = ClearanceItemSequence(item_class=model_items.Clearance)
        return self._clearances
    
  
    
   
    @property
    def materials(self):
        if self._materials is None:
            self._materials = MaterialItemSequence(item_class=model_items.Material)
        return self._materials
    
    

    @property
    def sources_ct(self):
        if self._sources_ct is None:
            self._sources_ct = NamedModelItemSequence(item_class=model_items.SourceCT)
        return self._sources_ct
            
    @property
    def sources_xray(self):
        if self._sources_xray is None:
            self._sources_xray = NamedModelItemSequence(item_class=model_items.SourceXray)
        return self._sources_xray
    
            
        
    def zero_origin(self):
        # old psp files had an origin option, origin is now always at (0, 0)
        
        origin = self.floorplan.geometry.origin_cm
        
        if origin[0] != 0 or origin[1] != 0:
            self.shift_cm(origin[0], origin[1])
            self.floorplan.geometry.origin_cm = [0, 0]
            
        
    def __str__(self):
        return yaml.dump(self.to_dict())
        
        
    def set_callbacks(self):

        self.shieldings.connect(self, self.shieldings.EVENT_UPDATE_ITEM,
                                self.update_shielding)
        
        self.clearances.connect(self, self.clearances.EVENT_UPDATE_ITEM,
                                self.update_clearance)
        
        self.materials.connect(self, self.materials.EVENT_UPDATE_ITEM,
                               self.update_material)
        
    def remove_item(self, item):
        container = self.sequence_for_item(item)
        container.remove(item)
        
    def update_material(self, event_data):
        # called when a material is changed. Check for change in name
        # and update the references for all walls
        self.logger.debug(f'Update material with event_data {str(event_data)}')
        item, label, old_value, value = event_data
        if label == labels.NAME:
            for shielding in self.shieldings:
                materials = shielding.materials
                update = False
                if len(materials) > 0 and old_value == materials[0][0]:
                    update = True
                    materials[0][0] = value
                if len(materials) > 1 and old_value == materials[1][0]:
                    update = True
                    materials[1][0] = value
                
                if update:
                    shielding.materials = materials
                
        
    def update_shielding(self, event_data):
        # called when a shielding is changed. Check for change in name
        # and update the references for all walls
        self.logger.debug(f'Update shielding with event_data {str(event_data)}')
        item, label, old_value, value = event_data

        
            
        if label == labels.NAME:
            # name of shielding changed!
            
            #update references in walls to ensure a consistent model
            for wall in self.walls:
                if wall.shielding == old_value:                    
                    wall.shielding = value
    
        # If shielding color or linewidth changed generate event for walls
        elif label == labels.COLOR:
            
            for wall in self.walls:
                if wall.shielding == item.name:
                    event_data = (wall, labels.COLOR, item.color, item.color)
                    wall.emit(wall.EVENT_UPDATE, event_data)
                    
        elif label == labels.LINEWIDTH:
         
            for wall in self.walls:
                if wall.shielding == item.name:
                    event_data = (wall, labels.LINEWIDTH, item.linewidth, item.linewidth)
                    wall.emit(wall.EVENT_UPDATE, event_data)
                    
    def update_clearance(self, event_data):
        self.logger.debug(f'Update shielding with event_data {str(event_data)}')
        item, label, old_value, value = event_data

        if label == labels.NAME:
            # name of shielding changed!
            
            #update references in sources nm to ensure a consistent model
            for item in self.sources_nm:
                if item.clearance == old_value:                    
                    item.clearance = value
                    
    def add_item(self, item):
        # add a item of any type used by gui
        sequence = self.sequence_for_item(item)
        sequence.append(item)
            
        
    def shift_cm(self, dx_cm, dy_cm):

         self._shift_walls_cm(dx_cm, dy_cm)
         self._shift_sources_cm(dx_cm, dy_cm)
         
    def _shift_walls_cm(self, shiftx, shifty):
         for wall in self.walls:
             vertices = copy.deepcopy(wall.vertices)
             vertices[0][0] += shiftx
             vertices[1][0] += shiftx
             vertices[0][1] += shifty
             vertices[1][1] += shifty
             
             self.logger.debug('Shift Wall form {wall.vertices}')
             wall.vertices = vertices
             
    def _shift_sources_cm(self, shiftx, shifty):
        
        
         containers = [self.critical_points, self.sources_nm,
                       self.sources_xray, self.sources_ct]
         
         for container in containers:
             for item in container:
                 position = item.position.copy()
                 position[0] += shiftx   
                 position[1] += shifty
                 item.position = position
                 
    def get_sequence_by_label(self, label):
        if label == labels.SOURCES_CT:
            return self.sources_ct
        elif label == labels.SOURCES_XRAY:
            return self.sources_xray
        elif label == labels.SOURCES_NM:
            return self.sources_nm
        elif label == labels.CRITICAL_POINTS:
            return self.critical_points
        elif label == labels.SHIELDINGS:
            return self.shieldings
        elif label == labels.WALLS:
            return self.walls
        elif label == labels.FLOORPLAN:
            return self.floorplan
        elif label == labels.DOSEMAP:
            return self.dosemap
        elif label == labels.CLEARANCE:
            return self.clearances
        elif label == labels.MATERIALS:
            return self.materials
        else:
            raise KeyError(label)

    def sequence_for_item(self, item):
        return self.get_sequence_by_label(item.label)
    
        
        
            
    def match_extent_to_floorplan(self):
        self.dosemap.extent = self.floorplan.extent
        
   
    # @property
    # def shielding_dict(self):
    #     return {s.name: s for s in self.shieldings}
        
    # def get_shielding_by_name(self, shielding_name):
    #     if len(self.shieldings) == 0: return None
    #     if shielding_name is None: return None
        
    #     shielding = [shielding for shielding in self.shieldings\
    #                      if shielding.name == shielding_name]
            
    #     if len(shielding) > 1:
    #         msg = f'Multiple shieldings exists with name: {shielding_name}'
    #         raise KeyError(msg)
    #     elif len(shielding) == 0:
    #         msg = f'No shieldings exists with name: {shielding_name}'
    #         raise KeyError(msg)
    #     return shielding[0]
    
    
    
    
    

    @classmethod
    def from_dict(cls, dct): 
        floorplan = Floorplan.from_dict(dct[labels.FLOORPLAN])

        dosemap_dct = dct.get(labels.DOSEMAP, None)
        
        if dosemap_dct is not None:
            dosemap = Dosemap.from_dict(dosemap_dct)
        else:
            dosemap = None
            
        material_lst = dct.get(labels.MATERIALS, None)
        if material_lst is not None:
            materials = [model_items.Material.from_dict(item) for item in material_lst]
        else:
            materials = None
            
        
        
        dosemap_style = DosemapStyle.from_dict(
                                        dct.get(labels.DOSEMAP_STYLE, None))
        
        
        
        shieldings = [model_items.Shielding.from_dict(item)\
                      for item in dct[labels.SHIELDINGS]]
            
            
        walls = [model_items.Wall.from_dict(item)\
                 for item in dct[labels.WALLS]]
            
            
        critical_points = [model_items.CriticalPoint.from_dict(item)\
                           for item in dct[labels.CRITICAL_POINTS]]
            
            
        sources_ct = [model_items.SourceCT.from_dict(item)\
                      for item in dct[labels.SOURCES_CT]]
            
        sources_nm = [model_items.SourceNM.from_dict(item)\
                      for item in dct[labels.SOURCES_NM]]

        sources_xray = [model_items.SourceXray.from_dict(item)\
                        for item in dct[labels.SOURCES_XRAY]]
            
        
            
            
        if labels.CLEARANCE in  dct.keys():
            clearances = ClearanceItemSequence.from_dict(dct)
        else:
            clearances = None
            
            
        
        model = cls(floorplan=floorplan, 
                    dosemap=dosemap, 
                    dosemap_style=dosemap_style, 
                    shieldings=shieldings, 
                    walls=walls, 
                    critical_points=critical_points, 
                    sources_ct=sources_ct, 
                    sources_nm=sources_nm, 
                    sources_xray=sources_xray, 
                    clearances=clearances,
                    materials=materials)

        return model
    
        
    def to_dict(self):
        floorplan = self.floorplan.to_dict()
        dosemap = self.dosemap.to_dict()
        dosemap_style = self.dosemap_style.to_dict()
        
    
        
        walls           = [wall.to_dict() for wall in self.walls]        
        critical_points = [item.to_dict() for item in self.critical_points]        
        sources_ct      = [source.to_dict() for source in self.sources_ct]
        sources_nm      = [source.to_dict() for source in self.sources_nm]
        sources_xray    = [source.to_dict() for source in self.sources_xray]
        
        # remove the default / empty item
        shieldings      = [item.to_dict() for item in self.shieldings[1:]]
        clearances      = [item.to_dict() for item in self.clearances[1:]]
        materials       = [item.to_dict() for item in self.materials[1:]]
        
            
        return {labels.FLOORPLAN:           floorplan,
                labels.DOSEMAP:             dosemap,
                labels.DOSEMAP_STYLE:       dosemap_style,
                labels.SHIELDINGS:          shieldings,
                labels.WALLS:               walls,
                labels.SOURCES_NM:          sources_nm,
                labels.SOURCES_XRAY:        sources_xray,
                labels.SOURCES_CT:          sources_ct,
                labels.CRITICAL_POINTS:     critical_points,
                labels.CLEARANCE:           clearances,
                labels.MATERIALS:           materials}
    
    


    def save_to_project_file(self, filename):    
        file, ext = os.path.splitext(filename)
        if ext.lower() == '.psp':
            filename = file + '.zip' # force saving old projects to ext zip
        
        self.filename = filename
                        
        project_dict = self.to_dict()
        image = project_dict[labels.FLOORPLAN].pop(labels.IMAGE)
        
        temp_yml = io.StringIO()
        yaml.dump(project_dict, temp_yml, default_flow_style=None)
        
        temp_img = io.BytesIO()
        imageio.imwrite(temp_img, image, format=".png")

        zf = zipfile.ZipFile(filename, "w")
        zf.writestr("project.yml", temp_yml.getvalue())
        zf.writestr("floorplan.png", temp_img.getvalue())
    
    
    @classmethod
    def load_from_project_file(cls, filename):
        try:
            with zipfile.ZipFile(filename) as zf:
                with zf.open("floorplan.png", "r") as f:
                    image = imageio.imread(f)
                    
                with zf.open("project.yml", "r") as f:
                    dct = yaml.safe_load(f)
                
                dct[labels.FLOORPLAN][labels.IMAGE] = image
                model = cls.from_dict(dct)
                
        
        except zipfile.BadZipFile:
            ### For backwards compatibility:
            if filename.endswith(".psp"):
                with open(filename, 'rb') as fp:
                    dct = pickle.load(fp)
                image = dct.pop("IMAGE_DATA")   
                dct[labels.FLOORPLAN][labels.IMAGE] = image
                dct[labels.FLOORPLAN].pop('Filename', None)
                model = cls.from_dict(dct)
                
            else:
                # IOError is picked up by GUI to show error dlg
                raise IOError(f'Could not read {filename}')
    
        model.filename = filename
        return model

class ModelItemSelector(Observable):
    model = None
    EVENT_SELECT_ITEM = 'event_select_item'
    _selected_item = None
    
    
    def __init__(self, model=None):
        Observable.__init__(self)
        self.connect_model(model)
        
    def disconnect_model(self):
        for container in (self.model.shieldings, 
                          self.model.walls, 
                          self.model.critical_points,
                          self.model.sources_nm, 
                          self.model.sources_ct, 
                          self.model.sources_xray,
                          self.model.clearances):
            
            container.disconnect(self)
    
    def connect_model(self, model):
        if self.model is not None:
            self.disconnect_model()
        
        self.model = model
        

        for container in (self.model.shieldings, 
                          self.model.walls, 
                          self.model.critical_points,
                          self.model.sources_nm, 
                          self.model.sources_ct, 
                          self.model.sources_xray,
                          self.model.clearances):
            
            container.connect(self, container.EVENT_ADD_ITEM, self.item_added)
            
            container.connect(self, container.EVENT_WILL_REMOVE_ITEM, 
                              self.will_remove_item)
    
    def item_added(self, item):
        # make new item the selected item always
        self.select_item(item)
        
        
    def will_remove_item(self, item):
        # called just before removal of item
        if self.selected_item is item:
            # change selected item to something sensible
            container = self.model.sequence_for_item(item)
            index = container.index(item)
            if len(container) > 1:
                if index == 0:
                    new_selected_item = container[1]
                else:
                    new_selected_item = container[index-1]
            else:
                new_selected_item = None 
            
            self.select_item(new_selected_item)
            
    def select_item(self, item):
        # if item is self._selected_item:
        #     return
        self._selected_item = item
        self.emit(self.EVENT_SELECT_ITEM, self.selected_item)   
        
        
        
    @property
    def selected_item(self):
        return self._selected_item



if __name__ == "__main__":
    # import yaml
    # with open('test_model_in.yml') as file:
    #     dct = yaml.safe_load(file)
    # model = Model.from_dict(dct)
    file = '../example_projects/LargeProject/project.zip'
    model = Model.load_from_project_file(file)
    



# VERSION: MAJOR.MINOR.BUGFIX
# When new functionality is added update minor. For new builds with only
# bugfixes etc update bugfix number.

# Pypi needs a new version number for each publication.

__version__ = '1.0.3'


import time
import numpy as np
import copy
from matplotlib.cm import get_cmap
from matplotlib.colors import LogNorm
from matplotlib.ticker import AutoMinorLocator, MultipleLocator
from pyrateshield.model_items import Wall, Shielding
from pyrateshield.gui.mpl_items_controller import PixelSizeLine
from pyrateshield.floorplan_and_dosemap import MeasuredGeometry
from pyrateshield.gui.mpl_view import PointClicker, MovePointByMouse, Mouse
from pyrateshield.gui.mpl_items_controller import MplItemsPlotter, MPLPixelSize
from pyrateshield.gui.mpl_picker import MPLPicker
from pyrateshield import labels, Observable, GLOBAL_LOG_LEVEL


LOG_LEVEL = GLOBAL_LOG_LEVEL
#LOG_LEVEL = 10


def plot_dosemap(ax, model, dosemap):
    extent = model.dosemap.extent        
    style = model.dosemap_style
    
    cmap = copy.copy(get_cmap(style.cmap_name))
    if style.alpha_gradient:
        alphas = np.linspace(0, 1, cmap.N)
        cmap._init()
        cmap._lut[:len(alphas), -1] = alphas            
    cmap.set_under(alpha=0)
    
    img = ax.imshow(dosemap, interpolation="bilinear")
    img.set_extent(extent)
    img.set_cmap(cmap)
    img.set_alpha(style.alpha)
    img.set_clim([style.vmin, style.vmax])
    img.set_norm(LogNorm(vmin=style.vmin, vmax=style.vmax))
            
    levels, colors, linestyles, linewidths, active = zip(*[cl for cl in style.contour_lines if cl[4]])
    CS = ax.contour(dosemap, extent=extent, levels=levels, 
                                colors=colors, linestyles=linestyles,
                                linewidths=linewidths, origin="upper")
    h,_ = CS.legend_elements()        
    leg = ax.legend(h, levels, handlelength=3, 
                                title="Contours [mSv]", framealpha=0.7)

    return img, CS, leg

            
class MplController(Observable):
    _dosemap_image = None
    _contour_lines = None
    _legend = None
    _floorplan_image = None
    
    _scale_line = None
    _origin_point = None

    _model = None
    lock_mouse = False
    
    _point_clicker = None
    
    EVENT_PICK_LEFT_CLICK = 'event_pick_left_click'
    EVENT_RIGHT_CLICK = 'event_right_click'
    EVENT_REFRESH = 'event_refresh'

    def __init__(self, dosemapper=None, view=None, model=None):       
        Observable.__init__(self, log_level=LOG_LEVEL)
        self.dosemapper = dosemapper
        self.view = view

        self.mouse = Mouse(self.view.canvas)
        
        self.mpl_items = MplItemsPlotter(model=model, axes=self.axes)
       
        self.mpl_picker = MPLPicker(mpl_items=self.mpl_items,
                                   item_selector=model.item_selector,
                                   axes = self.axes)
        self._enable_pick = True
        self.model = model
        
        self.set_callbacks()   
        
        self.refresh()
    
    
    def enable_picking(self):
        self._enable_pick = True
    
    def disable_picking(self):
        self._disable_pick = True
    
    
    @property
    def model(self):
       return self._model
   
    @model.setter
    def model(self, model):
       if self.model is model:
           return
       if self.model is not None:
           self.model.disconnect(self)
           self.model.item_selector.disconnect(self)
           
       self._model = model
       
       self.mpl_items.model = model       
     
       self.mpl_picker.item_selector = model.item_selector
       
       callback = self.select_item_callback
       self.model.item_selector.connect(self, 
                                    self.model.item_selector.EVENT_SELECT_ITEM, 
                                    callback)
              
       callback = self.new_floorplan
       
       self.model.connect(self,
                          self.model.EVENT_UPDATE_FLOORPLAN,
                          callback)
       self.refresh()

       
    def set_callbacks(self):
        # button callbacks
        button = self.view.buttons[self.view.dosemap_button_pyshield_label]
        button.clicked.connect(self.show_dosemap_pyshield)
        
        button = self.view.buttons[self.view.dosemap_button_radtracer_label]
        button.clicked.connect(self.show_dosemap_radtracer)
        
        button = self.view.buttons[self.view.refresh_button_label]
        button.clicked.connect(self.refresh)
        
        #self.axes.figure.canvas.mpl_connect('pick_event', self.pick_event)
        
        self.mouse.connect(self, self.mouse.LEFT_CLICK_EVENT, 
                           self.left_mouse_click)
        
        self.mouse.connect(self, self.mouse.RIGHT_CLICK_EVENT, 
                           self.right_mouse_click)
        
        
        event = self.view.toolbar.EVENT_TOOLBUTTON_CLICK
        self.view.toolbar.connect(self, event, self.toolbar_callback)
    
    
    def draw(self):        
        self.axes.figure.canvas.draw_idle()
    
    
    def left_mouse_click(self, position):
        self.logger.debug('Left Mouse Click')
        # check toolbar what to do
        toolname = self.view.toolbar.selected_tool
        
        self.logger.debug(f'Tool selected: {str(toolname)}')
        
        if toolname is None: # select, and drag items
            self.logger.debug('Start picking objects')
            # note vertex_index is None for point items
            model_item, vertex_index = self.mpl_picker.pick_model_item(position)
            
            if model_item is None:
                self.logger.debug('Nothing close to the mouse position')
                # nothing selected random click
                return
            self.logger.debug(f'Item of type {type(model_item)} picked')
                
            if model_item is self.model.item_selector.selected_item:
                if isinstance(model_item, Wall):
                    if vertex_index is not None:
                        # selected item clicked, start mouse interaction / drag
                        MovePointByMouse(mouse=self.mouse, 
                                         model_item=model_item,
                                         vertex_index=vertex_index)
                else:
                    MovePointByMouse(mouse=self.mouse, model_item=model_item)
                                     
            else:
                # select picked model_item, make active
                self.model.item_selector.select_item(model_item)
        
        
        elif toolname in (labels.SOURCES_CT, labels.SOURCES_NM, 
                          labels.SOURCES_XRAY, labels.CRITICAL_POINTS):
            # create point item
            self.add_model_item(label=toolname, position=position)
        elif toolname == labels.WALLS:
            self.add_wall_by_mouse(position)
        elif toolname == labels.FLOORPLAN:
            self.move_items(position)
    
    
    def right_mouse_click(self, position):
        # check toolbar what to do
        toolname = self.view.toolbar.selected_tool
        
        self.logger.debug(f'Position!: {position}')
        if toolname is None:
            model_item, vertex_index = self.mpl_picker.pick_model_item(position)
            event_data = (model_item, vertex_index, position)
            # note model item can be None if no item was in radius
            self.emit(self.EVENT_RIGHT_CLICK, event_data)
 
 
    @property
    def floorplan_image(self):
        if self._floorplan_image is None:
            self._floorplan_image = self.axes.imshow(((0,0), (0,0)))
        return self._floorplan_image
        
        
    def select_item_callback(self, item):
        self.toggle_delete()


    def toggle_delete(self):
        if self.model.item_selector.selected_item is not None:
            self.view.toolbar.button_enabled('delete', True)
        else:
            self.view.toolbar.button_enabled('delete', False)
        
        
    def clear(self):
        self.axes.cla()
        self._floorplan_image = None
        self.mpl_items.clear()
        
        
    @property
    def axes(self):
        return self.view.canvas.ax
        
        
    def refresh(self, _=None):
        self.clear()
        self.model.floorplan.disconnect(self)
        self.new_floorplan()
        self.mpl_items.refresh()
        #self.mpl_items_selector.refresh()
        self.toggle_delete()
        self.draw()


    def update_floorplan(self, event_data=None):
        model_item, label, old_value, value = event_data
        if label == labels.IMAGE:
            self.floorplan_image.set_data(self.model.floorplan.image)
            self.model.match_extent_to_floorplan()
            self.refresh()
            
        elif label == labels.GEOMETRY:
            self.floorplan_image.set_extent(self.model.floorplan.extent)
        self.draw()
        #self.plot_floorplan()
    
    
    def new_floorplan(self, _=None):
        self.logger.debug('Plotting new floorplan')
        self.floorplan_image.set_data(self.model.floorplan.image)
        self.floorplan_image.set_extent(self.model.floorplan.extent)
        
        self.axes.set_xlim(self.model.dosemap.extent[:2])
        self.axes.set_ylim(self.model.dosemap.extent[2:])
                
        # callback for changes in geometry
        callback = self.update_floorplan
        self.model.floorplan.connect(self, 
                                     self.model.floorplan.EVENT_UPDATE,
                                     callback)
        
        self.draw()
    
   
    def get_dosemap(self):
        start = time.time()
        self.model.dosemap.extent = self.axes.get_xlim() + self.axes.get_ylim()
        dosemap = self.dosemapper.get_dosemap(self.model)          
        end = time.time()        
        exec_time = round(end-start, 2)
        self.logger.info(f'Dosemap calculated in {exec_time}s')
        return dosemap
    
    
    def show_dosemap_pyshield(self):
        engine = self.model.dosemap.engine
        self.model.dosemap.engine = labels.PYSHIELD
        self.show_dosemap()
        self.model.dosemap.engine = engine
    
    
    def show_dosemap_radtracer(self):
        engine = self.model.dosemap.engine
        self.model.dosemap.engine = labels.RADTRACER
        self.show_dosemap()
        self.model.dosemap.engine = engine
    
    
    def show_dosemap(self):
        # Remove all dosemap related elements from the current axes
        clear_elements = [self._dosemap_image, self._legend]
        if self._contour_lines is not None:
            clear_elements += self._contour_lines.collections
        
        ax_elements = self.axes.get_children()
        for element in clear_elements:
            if element in ax_elements:
                element.remove()
        
        # Calculate new dosemap and plot elements
        dosemap = self.get_dosemap()
        if dosemap is not None:            
            img, CS, leg = plot_dosemap(self.axes, self.model, dosemap)            
            self._dosemap_image = img
            self._contour_lines = CS
            self._legend = leg
            
        self.draw()


    def add_model_item(self, label, position):
        model_items = self.model.get_sequence_by_label(label)
        new_item = model_items.add_new_item(position=position)
        
        self.model.item_selector.select_item(new_item)
        
        self.view.toolbar.select_checkable_tool(toolname=None)
    
    
    def add_wall_by_mouse(self, position=None):       

        vertices = [position.copy(), position.copy()]
        
        # Find active shielding
        if isinstance(self.model.item_selector.selected_item, PixelSizeLine):
            shielding = labels.EMPTY_SHIELDING
        elif isinstance(self.model.item_selector.selected_item, Wall):
            shielding = self.model.item_selector.selected_item.shielding
        elif isinstance(self.model.item_selector.selected_item, Shielding):
            shielding = self.model.item_selector.selected_item.name
        else:
            shielding = labels.EMPTY_SHIELDING
            
            
        # Create new empyt wall
        new_item = self.model.walls.add_new_item(vertices=vertices,
                                                 shielding=shielding)
        
        self.model.item_selector.select_item(new_item)
        
        # Move second vertex by mouse
        def finished(_=None):
            self.view.toolbar.select_checkable_tool(toolname=None)
        
        mover = MovePointByMouse(mouse=self.mouse, model_item=new_item,
                                 vertex_index=1, hold=True)
        
        mover.connect(self, mover.EVENT_FINISHED, finished)
        
        #ugly fix for small displacement of first vertex 
        new_item.set_vertex(0, [*position])
        
        self.draw()

    
    def move_items(self, event_data):
        self.temp_data = None
        self.logger.debug('Start Moving!')
        def move_model(position):
            if self.temp_data is None:
                self.temp_data = self.mouse.click_position
            
            dx_cm =  position[0] - self.temp_data[0]
            dy_cm = position[1] - self.temp_data[1] 
            
            self.logger.debug(f'Shifting {dx_cm}, {dy_cm}')
            
            self.model.shift_cm(dx_cm, dy_cm)
            self.temp_data = position
        
        def release(_):
            self.mouse.disconnect(self, self.mouse.MOVE_EVENT)
            self.mouse.disconnect(self, self.mouse.RELEASE_EVENT)
        
        
        self.mouse.connect(self, self.mouse.MOVE_EVENT, move_model)
        self.mouse.connect(self, self.mouse.RELEASE_EVENT, release)


    def toolbar_callback(self, event_data):
        toolname, checked = event_data
        
        if toolname == labels.PYSHIELD:
            self.show_dosemap_pyshield()
        elif toolname == labels.RADTRACER:
            self.show_dosemap_radtracer()
        elif toolname == labels.FLOORPLAN:
            self.move_items(event_data)
        elif toolname == 'refresh':
            self.refresh()
        elif toolname == 'delete':
            if self.model.item_selector.selected_item is not None:
                self.model.remove_item(self.model.item_selector.selected_item)




import os
import pandas as pd


#ENERGY_keV = 'Energy [keV]'
MFP = 'mfp'
MASS_ATTENUATION = 'mu/p [cm^2/g]'
ENERGY_MeV = 'Energy [MeV]'

folder = os.path.dirname(__file__)

file = os.path.join(folder, 'buildup.xls')
BUILDUP_TABLES = pd.read_excel(os.path.join(folder, file), sheet_name=None)

file = os.path.join(folder, 'attenuation.xls')
ATTENUATION_TABLES = pd.read_excel(os.path.join(folder, file), sheet_name=None)





  